from pathlib import Path
from typing import Optional
from unittest.mock import MagicMock, patch

import dask.array as da
import numpy as np
import pandas as pd
import pytest
import soundfile as sf
from matplotlib.axes import Axes

# テスト対象のクラスをインポート
from wandas.frames.channel import ChannelFrame
from wandas.frames.spectrogram import SpectrogramFrame
from wandas.utils.frame_dataset import (
    ChannelFrameDataset,
    FrameDataset,
    LazyFrame,  # 新しいクラスをインポート
    SpectrogramFrameDataset,
    _SampledFrameDataset,
)
from wandas.utils.types import NDArrayReal

_da_from_array = da.from_array  # type: ignore [unused-ignore]

# --- Test Fixtures ---


@pytest.fixture  # type: ignore [misc, unused-ignore]
def sample_wav_data() -> tuple[int, NDArrayReal]:
    """Generate sample WAV data."""
    sampling_rate = 16000
    duration = 1.0
    n_samples = int(sampling_rate * duration)
    # Create stereo data (2 channels)
    data = np.random.uniform(-0.5, 0.5, size=(n_samples, 2)).astype(np.float32)
    return sampling_rate, data


@pytest.fixture  # type: ignore [misc, unused-ignore]
def sample_csv_data() -> tuple[int, pd.DataFrame]:
    """Generate sample CSV data."""
    sampling_rate = 100
    duration = 1.0
    n_samples = int(sampling_rate * duration)
    time_col = np.linspace(0, duration, n_samples, endpoint=False)
    ch1_data = np.sin(2 * np.pi * 5 * time_col)  # 5 Hz sine wave
    ch2_data = np.cos(2 * np.pi * 10 * time_col)  # 10 Hz cosine wave
    df = pd.DataFrame({"time": time_col, "SensorA": ch1_data, "SensorB": ch2_data})
    return sampling_rate, df


@pytest.fixture  # type: ignore [misc, unused-ignore]
def create_test_files(
    tmp_path: Path,
    sample_wav_data: tuple[int, NDArrayReal],
    sample_csv_data: tuple[int, pd.DataFrame],
) -> Path:
    """Create sample WAV and CSV files in a temporary directory."""
    wav_sr, wav_data = sample_wav_data
    csv_sr, csv_df = sample_csv_data

    # Create WAV files
    sf.write(tmp_path / "test1.wav", wav_data, wav_sr)
    sf.write(tmp_path / "test2.wav", wav_data * 0.5, wav_sr)  # Different data

    # Create CSV file
    csv_df.to_csv(tmp_path / "test3.csv", index=False)

    # Create a subdirectory for recursive test
    sub_dir = tmp_path / "subdir"
    sub_dir.mkdir()
    sf.write(sub_dir / "sub_test.wav", wav_data * 0.8, wav_sr)

    # Create a non-target file
    (tmp_path / "other.txt").write_text("ignore me")

    # Create an empty subdirectory for empty-dataset tests
    empty_subdir = tmp_path / "empty_subdir"
    empty_subdir.mkdir()

    return tmp_path


# --- Test LazyFrame ---


class TestLazyFrame:
    def test_lazy_frame_init(self) -> None:
        """Test LazyFrame initialization."""
        file_path = Path("/path/to/file.wav")
        lazy_frame: LazyFrame[ChannelFrame] = LazyFrame(file_path)

        assert lazy_frame.file_path == file_path
        assert lazy_frame.frame is None
        assert lazy_frame.is_loaded is False
        assert lazy_frame.load_attempted is False

    def test_ensure_loaded(self) -> None:
        """Test LazyFrame.ensure_loaded method."""
        file_path = Path("/path/to/file.wav")
        lazy_frame: LazyFrame[ChannelFrame] = LazyFrame(file_path)

        # モックローダー関数
        mock_frame = ChannelFrame.from_ndarray(
            np.array([[0.1, 0.2], [0.3, 0.4]]), 44100
        )
        loader = MagicMock(return_value=mock_frame)

        # 初回ロード
        result = lazy_frame.ensure_loaded(loader)

        assert result is mock_frame
        assert lazy_frame.frame is mock_frame
        assert lazy_frame.is_loaded is True
        assert lazy_frame.load_attempted is True
        loader.assert_called_once_with(file_path)

        # 2回目のロード（キャッシュから取得）
        loader.reset_mock()
        cached_result = lazy_frame.ensure_loaded(loader)

        assert cached_result is mock_frame
        loader.assert_not_called()

    def test_ensure_loaded_failure(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test LazyFrame.ensure_loaded with loader failure."""
        file_path = Path("/path/to/file.wav")
        lazy_frame: LazyFrame[ChannelFrame] = LazyFrame(file_path)

        # 失敗するローダー関数
        error_msg = "Failed to load file"
        loader = MagicMock(side_effect=RuntimeError(error_msg))

        # ロード試行
        result = lazy_frame.ensure_loaded(loader)

        assert result is None
        assert lazy_frame.frame is None
        assert lazy_frame.is_loaded is True  # ロードは試みた
        assert lazy_frame.load_attempted is True
        loader.assert_called_once_with(file_path)

        # ログ確認
        assert any(
            "Failed to load file" in record.message and record.levelname == "ERROR"
            for record in caplog.records
        )

    def test_reset(self) -> None:
        """Test LazyFrame.reset method."""
        file_path = Path("/path/to/file.wav")
        lazy_frame: LazyFrame[ChannelFrame] = LazyFrame(file_path)

        # フレームをロード
        mock_frame = ChannelFrame.from_ndarray(
            np.array([[0.1, 0.2], [0.3, 0.4]]), 44100
        )
        loader = MagicMock(return_value=mock_frame)
        lazy_frame.ensure_loaded(loader)

        # リセット前の状態確認
        assert lazy_frame.frame is mock_frame
        assert lazy_frame.is_loaded is True
        assert lazy_frame.load_attempted is True

        # リセット
        lazy_frame.reset()

        # リセット後の状態確認
        assert lazy_frame.frame is None
        assert lazy_frame.is_loaded is False
        assert lazy_frame.load_attempted is False


# --- Test FrameDataset (Abstract Base Class) ---


class TestFrameDatasetABC:
    def test_frame_dataset_cannot_instantiate(self) -> None:
        """Verify FrameDataset ABC cannot be instantiated directly."""
        with pytest.raises(TypeError, match="Can't instantiate abstract class"):
            FrameDataset("/some/path")  # type: ignore [abstract]  # 抽象クラスは直接インスタンス化できない

    def test_abstract_methods_must_be_implemented(self, tmp_path: Path) -> None:
        """Test that abstract methods must be implemented in subclasses."""

        # テスト用のサブクラスを作成し、抽象メソッドを実装していない
        class IncompleteFrameDataset(FrameDataset[ChannelFrame]):
            pass

        # Pythonのバージョンによって例外メッセージが違うので、部分一致で確認
        with pytest.raises(TypeError, match="abstract class"):
            IncompleteFrameDataset(str(tmp_path))  # type: ignore [abstract]

        # テスト用のサブクラスを作成し、抽象メソッドを実装
        class MinimalFrameDataset(FrameDataset[ChannelFrame]):
            def _load_file(self, file_path: Path) -> Optional[ChannelFrame]:
                return ChannelFrame.from_ndarray(np.zeros((10, 2)), 44100)

        # これは例外を投げない
        try:
            MinimalFrameDataset(str(tmp_path), lazy_loading=True)
        except TypeError:
            pytest.fail("正しく実装されたサブクラスがインスタンス化に失敗しました")


# --- Test ChannelFrameDataset ---


class TestChannelFrameDataset:
    def test_init_lazy(self, create_test_files: Path) -> None:
        """Test lazy initialization."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(
            str(folder_path), recursive=False, lazy_loading=True
        )

        assert dataset._lazy_loading is True
        assert len(dataset) == 3  # test1.wav, test2.wav, test3.csv
        # LazyFrameの確認 - すべてロードされていない
        assert len(dataset._lazy_frames) == 3
        assert all(not lf.is_loaded for lf in dataset._lazy_frames)
        assert all(lf.frame is None for lf in dataset._lazy_frames)
        assert dataset.folder_path == folder_path
        assert dataset.file_extensions == [
            ".wav",
            ".mp3",
            ".flac",
            ".csv",
        ]  # Default from from_folder via __init__
        assert dataset._recursive is False

    def test_init_eager(self, create_test_files: Path) -> None:
        """Test eager initialization."""
        folder_path = create_test_files

        # Initialize with eager loading
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=False)

        assert dataset._lazy_loading is False
        assert len(dataset) == 3

        # Verify all frames are loaded (not None)
        assert all(lf.is_loaded for lf in dataset._lazy_frames)
        assert all(lf.frame is not None for lf in dataset._lazy_frames)

        # Verify all frames are instances of ChannelFrame
        assert all(isinstance(lf.frame, ChannelFrame) for lf in dataset._lazy_frames)

        # Verify frame labels match the expected file names
        file_stems = [p.stem for p in dataset._get_file_paths()]
        frame_labels = [
            lf.frame.label for lf in dataset._lazy_frames if lf.frame is not None
        ]
        assert frame_labels == file_stems

    def test_init_recursive(self, create_test_files: Path) -> None:
        """Test recursive file discovery."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(
            str(folder_path), recursive=True, lazy_loading=True
        )
        assert len(dataset) == 4  # Includes subdir/sub_test.wav
        assert dataset._recursive is True
        assert Path("subdir/sub_test.wav") in [
            p.relative_to(folder_path) for p in dataset._get_file_paths()
        ]

    def test_init_custom_extensions(self, create_test_files: Path) -> None:
        """Test custom file extensions."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(
            str(folder_path), file_extensions=[".csv"], lazy_loading=True
        )
        assert len(dataset) == 1
        assert dataset._lazy_frames[0].file_path.name == "test3.csv"
        assert dataset.file_extensions == [".csv"]

    def test_init_no_files(self, tmp_path: Path) -> None:
        """Test initialization with a folder containing no matching files."""
        dataset = ChannelFrameDataset(str(tmp_path), lazy_loading=True)
        assert len(dataset) == 0
        assert len(dataset._lazy_frames) == 0

    def test_init_folder_not_found(self) -> None:
        """Test initialization with a non-existent folder."""
        with pytest.raises(FileNotFoundError):
            ChannelFrameDataset("non_existent_folder")

    def test_len(self, create_test_files: Path) -> None:
        """Test __len__ method."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        assert len(dataset) == 3

    def test_getitem_lazy(self, create_test_files: Path) -> None:
        """Test __getitem__ with lazy loading."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)

        # Access first item - should trigger load
        with patch.object(dataset, "_load_file", wraps=dataset._load_file) as mock_load:
            frame0 = dataset[0]
            assert isinstance(frame0, ChannelFrame)
            assert frame0.label == "test1"  # From filename stem
            assert dataset._lazy_frames[0].is_loaded is True
            assert dataset._lazy_frames[0].frame is frame0
            mock_load.assert_called_once_with(dataset._lazy_frames[0].file_path)

        # Access again - should use cache
        with patch.object(ChannelFrameDataset, "_load_file") as mock_load_again:
            frame0_cached = dataset[0]
            assert frame0_cached is frame0
            mock_load_again.assert_not_called()

    def test_getitem_index_error(self, create_test_files: Path) -> None:
        """Test __getitem__ with out-of-bounds index."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        with pytest.raises(IndexError):
            _ = dataset[3]
        with pytest.raises(IndexError):
            _ = dataset[-1]  # Negative indexing not supported by this logic

    def test_load_file_wav(self, create_test_files: Path) -> None:
        """Test _load_file method for WAV."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        wav_path = dataset._lazy_frames[0].file_path  # test1.wav

        frame = dataset._load_file(wav_path)
        assert isinstance(frame, ChannelFrame)
        assert frame.label == "test1"
        assert frame.sampling_rate == 16000  # From sample_wav_data
        assert frame.n_channels == 2
        assert isinstance(frame.data, np.ndarray)

    def test_load_file_csv(self, create_test_files: Path) -> None:
        """Test _load_file method for CSV."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        csv_path = dataset._lazy_frames[2].file_path  # test3.csv

        # Need to explicitly call _load_file as from_file handles CSV specifics
        # We simulate this by calling ChannelFrame.from_file directly here
        frame = ChannelFrame.from_file(csv_path, time_column="time")

        assert isinstance(frame, ChannelFrame)
        assert frame.label == "test3"
        assert frame.sampling_rate == 100  # From sample_csv_data
        assert frame.n_channels == 2  # SensorA, SensorB
        assert frame.labels == ["SensorA", "SensorB"]
        assert isinstance(frame.data, np.ndarray)

    def test_load_file_resampling(
        self, create_test_files: Path, sample_wav_data: tuple[int, NDArrayReal]
    ) -> None:
        """Test _load_file triggers resampling when SR mismatches."""
        folder_path = create_test_files
        original_sr, _ = sample_wav_data  # Should be 16000
        target_sr = 8000
        dataset = ChannelFrameDataset(
            str(folder_path), sampling_rate=target_sr, lazy_loading=True
        )
        wav_path = dataset._lazy_frames[0].file_path

        # Mock ChannelFrame.resampling to check if it's called
        # We don't need the return value if we only check the call itself.
        with patch.object(ChannelFrame, "resampling", autospec=True) as mock_resample:
            # We still need to call _load_file to trigger the resampling call
            try:
                # _load_file might raise an error if the mocked resampling
                # doesn't return something usable by subsequent code within _load_file,
                # but for just checking the call, this might be sufficient.
                # Alternatively, set a minimal return_value if needed.
                mock_resample.return_value = MagicMock()  # Provide a basic mock return
                _ = dataset._load_file(wav_path)
            except Exception:
                # Ignore errors after the mock call if only checking the call
                pass

            # Check if resampling was actually called with the correct target SR
            mock_resample.assert_called_once()
            # Verify the keyword argument explicitly
            args, kwargs = mock_resample.call_args
            assert "target_sr" in kwargs
            assert kwargs["target_sr"] == target_sr
            # Optionally check the instance it was called on had the original SR
            assert isinstance(args[0], ChannelFrame)
            assert args[0].sampling_rate == original_sr

    def test_load_file_error(
        self, tmp_path: Path, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test error handling during file loading."""
        folder_path = tmp_path
        # Create a corrupted/invalid file
        invalid_file = folder_path / "invalid.wav"
        invalid_file.write_text("this is not a wav file")

        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        result = dataset._ensure_loaded(0)
        assert result is None
        # Check logs for warning from _ensure_loaded
        assert any(
            "Failed to load or initialize file" in record.message
            and record.levelname == "ERROR"
            for record in caplog.records
        )
        # Ensure the frame is marked as None after failure
        assert dataset._lazy_frames[0].frame is None
        assert dataset._lazy_frames[0].is_loaded is True  # ロードは試みた

        # Test _load_all_files catching errors
        caplog.clear()
        dataset_eager = ChannelFrameDataset(str(folder_path), lazy_loading=False)
        assert (
            dataset_eager._lazy_frames[0].frame is None
        )  # Should be None due to load failure
        # 新しい実装では、エラーログはERRORレベルで出力される
        assert any(
            "Failed to load or initialize file" in record.message
            for record in caplog.records
        )

    def test_apply_lazy(self, create_test_files: Path) -> None:
        """Test lazy application of a function."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)

        def dummy_transform(frame: ChannelFrame) -> ChannelFrame:
            # Simple transform: reverse data
            return frame._create_new_instance(data=frame._data[:, ::-1])

        transformed_dataset = dataset.apply(dummy_transform)

        assert isinstance(transformed_dataset, ChannelFrameDataset)
        assert transformed_dataset._lazy_loading is True
        assert transformed_dataset._source_dataset is dataset
        assert transformed_dataset._transform is dummy_transform
        assert len(transformed_dataset) == len(dataset)
        assert all(
            not lf.is_loaded for lf in transformed_dataset._lazy_frames
        )  # Still lazy

        # Access item to trigger transformation
        original_frame0 = dataset[0]  # Load original
        transformed_frame0 = transformed_dataset[0]  # Load and transform

        assert isinstance(transformed_frame0, ChannelFrame)
        assert transformed_dataset._lazy_frames[0].is_loaded is True
        assert transformed_dataset._lazy_frames[0].frame is transformed_frame0
        # Verify transformation was applied (compare computed data)
        assert transformed_frame0 is not None
        assert original_frame0 is not None
        np.testing.assert_allclose(
            transformed_frame0.compute(),
            original_frame0.compute()[:, ::-1],
        )

        # Access again, should use cache
        with patch.object(
            transformed_dataset, "_transform", wraps=transformed_dataset._transform
        ) as mock_transform:
            cached_transformed = transformed_dataset[0]
            assert cached_transformed is transformed_frame0
            mock_transform.assert_not_called()

    def test_apply_chaining(self, create_test_files: Path) -> None:
        """Test chaining apply calls."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)

        def transform1(frame: ChannelFrame) -> ChannelFrame:
            return frame._create_new_instance(data=frame._data * 2)

        def transform2(frame: ChannelFrame) -> ChannelFrame:
            return frame._create_new_instance(data=frame._data + 1)

        chained_dataset = dataset.apply(transform1).apply(transform2)

        assert isinstance(chained_dataset, ChannelFrameDataset)
        assert chained_dataset._lazy_loading is True
        assert chained_dataset._transform is transform2
        assert chained_dataset._source_dataset is not None
        source_ds = chained_dataset._source_dataset
        assert source_ds._transform is transform1
        assert source_ds._source_dataset is dataset

        # Trigger computation
        original_frame0 = dataset[0]
        final_frame0 = chained_dataset[0]

        assert original_frame0 is not None
        assert final_frame0 is not None

        expected_data = (original_frame0.compute() * 2) + 1
        np.testing.assert_allclose(final_frame0.compute(), expected_data)

    def test_apply_error_in_transform(
        self, create_test_files: Path, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test error handling when the applied function fails."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)

        def failing_transform(frame: ChannelFrame) -> ChannelFrame:
            raise ValueError("Transform failed!")

        transformed_dataset = dataset.apply(failing_transform)

        # 新しい実装では例外はキャッチされてNoneが返される
        result = transformed_dataset[0]
        assert result is None

        # Check logs for warning
        assert any(
            "Transform failed!" in record.message and record.levelname == "WARNING"
            for record in caplog.records
        )
        # Ensure frame is marked as is_loaded but with frame=None
        assert transformed_dataset._lazy_frames[0].frame is None
        assert transformed_dataset._lazy_frames[0].is_loaded is True

    def test_resample_trim_normalize(self, create_test_files: Path) -> None:
        """Test resample, trim, normalize methods (which use apply)."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        target_sr = 8000

        # Resample
        resampled_ds = dataset.resample(target_sr)
        assert isinstance(resampled_ds, ChannelFrameDataset)
        assert resampled_ds._lazy_loading is True
        assert resampled_ds._source_dataset is dataset
        assert resampled_ds._transform is not None
        # Trigger and check SR
        resampled_frame0 = resampled_ds[0]
        assert resampled_frame0 is not None
        assert resampled_frame0.sampling_rate == target_sr

        # Trim (apply on resampled)
        trimmed_ds = resampled_ds.trim(start=0.1, end=0.5)
        assert isinstance(trimmed_ds, ChannelFrameDataset)
        assert trimmed_ds._lazy_loading is True
        assert trimmed_ds._source_dataset is resampled_ds
        # Trigger and check duration/samples
        trimmed_frame0 = trimmed_ds[0]
        assert trimmed_frame0 is not None
        expected_duration = 0.5 - 0.1
        expected_samples = int(expected_duration * target_sr)
        assert trimmed_frame0.duration == pytest.approx(expected_duration)
        assert trimmed_frame0.n_samples == expected_samples

    def test_stft_lazy(self, create_test_files: Path) -> None:
        """Test lazy STFT using the ChannelFrame.stft result as ground truth."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        n_fft = 512
        hop_length = 128

        # create the SpectrogramFrameDataset lazily
        stft_dataset = dataset.stft(n_fft=n_fft, hop_length=hop_length)
        assert isinstance(stft_dataset, SpectrogramFrameDataset)
        assert stft_dataset._lazy_loading is True
        assert stft_dataset._source_dataset is dataset
        assert len(stft_dataset) == len(dataset)
        assert all(not lf.is_loaded for lf in stft_dataset._lazy_frames)

        # compute expected spectrogram from the already‐loaded ChannelFrame
        original_frame = dataset[0]
        assert original_frame is not None
        expected_spec = original_frame.stft(n_fft=n_fft, hop_length=hop_length)

        # now load the spectrogram from the dataset under test
        stft_frame = stft_dataset[0]
        assert isinstance(stft_frame, SpectrogramFrame)
        assert stft_frame.n_fft == n_fft
        assert stft_frame.hop_length == hop_length
        assert stft_frame.sampling_rate == original_frame.sampling_rate

        # compare shapes
        assert stft_frame.shape == expected_spec.shape

        # compare the actual complex STFT values
        np.testing.assert_allclose(
            stft_frame.compute(), expected_spec.compute(), atol=1e-6
        )

        # second access must hit the cache and not retrigger transform
        with patch.object(
            stft_dataset, "_transform", wraps=stft_dataset._transform
        ) as mock_transform:
            cached = stft_dataset[0]
            assert cached is stft_frame
            mock_transform.assert_not_called()

    def test_sample(self, create_test_files: Path) -> None:
        """Test sampling the dataset."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        n_total = len(dataset)

        # Sample by number
        n_sample = 2
        sampled_n = dataset.sample(n=n_sample, seed=42)
        assert isinstance(sampled_n, _SampledFrameDataset)
        assert len(sampled_n) == n_sample
        assert sampled_n._original_dataset is dataset
        # Check sampled indices are subset of original
        original_paths = [p.name for p in dataset._get_file_paths()]
        sampled_paths = [lf.file_path.name for lf in sampled_n._lazy_frames]
        assert all(p in original_paths for p in sampled_paths)

        # Sample by ratio
        ratio = 0.5
        # Expect ceil(3 * 0.5) = 2 samples
        sampled_r = dataset.sample(ratio=ratio, seed=42)
        assert isinstance(sampled_r, _SampledFrameDataset)
        assert len(sampled_r) == int(n_total * ratio)
        assert sampled_r._original_dataset is dataset

        # Sample default (10% or min 1)
        sampled_def = dataset.sample(seed=42)
        assert len(sampled_def) == max(1, int(n_total * 0.1))
        assert sampled_def._original_dataset is dataset  # type: ignore[attr-defined]

        # Sample all
        sampled_all = dataset.sample(n=n_total)
        assert len(sampled_all) == n_total
        assert sampled_all._original_dataset is dataset  # type: ignore[attr-defined]

        # Sample more than available
        sampled_more = dataset.sample(n=n_total + 1)
        assert len(sampled_more) == n_total
        assert sampled_more._original_dataset is dataset  # type: ignore[attr-defined]

        # Sample from empty dataset
        empty_ds = ChannelFrameDataset(
            str(folder_path / "empty_subdir"), lazy_loading=True
        )
        sampled_empty = empty_ds.sample(n=1)
        assert len(sampled_empty) == 0

        # Check lazy loading is preserved in sampled dataset
        sampled_lazy = dataset.sample(n=1, seed=1)
        assert sampled_lazy._lazy_loading is True
        assert all(not lf.is_loaded for lf in sampled_lazy._lazy_frames)

        # Access item in sampled dataset
        original_index = sampled_lazy._original_indices[0]  # type: ignore[attr-defined]
        original_frame = dataset[original_index]  # Ensure original is loaded
        sampled_frame = sampled_lazy[0]  # Trigger load via sampled dataset
        assert isinstance(sampled_frame, ChannelFrame)
        assert original_frame is not None
        assert sampled_frame is not None
        assert sampled_frame.label == original_frame.label
        np.testing.assert_array_equal(sampled_frame.compute(), original_frame.compute())
        assert sampled_lazy._lazy_frames[0].is_loaded is True

        # Check apply on sampled dataset
        def gain_transform(f: ChannelFrame) -> ChannelFrame:
            return f * 2

        transformed_sampled = sampled_lazy.apply(gain_transform)
        assert isinstance(transformed_sampled, _SampledFrameDataset)
        assert transformed_sampled._original_dataset._transform is gain_transform

        # Trigger computation
        final_frame = transformed_sampled[0]
        assert final_frame is not None
        assert original_frame is not None
        expected_data = original_frame.compute() * 2
        np.testing.assert_allclose(final_frame.compute(), expected_data)

    def test_get_metadata(self, create_test_files: Path) -> None:
        """Test get_metadata method."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(
            str(folder_path), lazy_loading=True, sampling_rate=8000
        )

        meta_lazy = dataset.get_metadata()
        assert meta_lazy["folder_path"] == str(folder_path)
        assert meta_lazy["file_count"] == 3
        assert meta_lazy["loaded_count"] == 0
        assert meta_lazy["target_sampling_rate"] == 8000
        # Actual SR might be different until loaded, or None if load fails
        assert (
            meta_lazy["actual_sampling_rate"] == 8000
        )  # Inherits target if not loaded
        assert meta_lazy["lazy_loading"] is True
        assert meta_lazy["has_transform"] is False
        assert meta_lazy["frame_type"] == "Unknown"  # Not loaded yet

        # Load one item
        _ = dataset[0]
        meta_loaded_one = dataset.get_metadata()
        assert meta_loaded_one["loaded_count"] == 1
        assert (
            meta_loaded_one["actual_sampling_rate"] == 8000
        )  # Now reflects loaded frame (after resampling)
        assert meta_loaded_one["frame_type"] == "ChannelFrame"

        # Test with transformation
        transformed_ds = dataset.apply(lambda f: f * 2)
        meta_transformed = transformed_ds.get_metadata()
        assert meta_transformed["has_transform"] is True
        assert (
            meta_transformed["loaded_count"] == 0
        )  # Transform dataset has its own cache
        assert meta_transformed["frame_type"] == "Unknown"

        # Load item in transformed dataset
        _ = transformed_ds[0]
        meta_transformed_loaded = transformed_ds.get_metadata()
        assert meta_transformed_loaded["loaded_count"] == 1
        assert (
            meta_transformed_loaded["frame_type"] == "ChannelFrame"
        )  # Type after transform

        # Test with eager loading
        dataset_eager = ChannelFrameDataset(str(folder_path), lazy_loading=False)
        meta_eager = dataset_eager.get_metadata()
        assert meta_eager["lazy_loading"] is False
        assert meta_eager["loaded_count"] == 3
        assert meta_eager["frame_type"] == "ChannelFrame"

    def test_from_folder(self, create_test_files: Path) -> None:
        """Test the from_folder class method."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset.from_folder(
            str(folder_path),
            sampling_rate=8000,
            file_extensions=[".wav"],
            recursive=True,
            lazy_loading=False,
        )
        assert isinstance(dataset, ChannelFrameDataset)
        assert dataset.sampling_rate == 8000
        assert dataset.file_extensions == [".wav"]
        assert dataset._recursive is True
        assert dataset._lazy_loading is False
        assert len(dataset) == 3  # Only wav files, including subdir
        assert all(isinstance(lf.frame, ChannelFrame) for lf in dataset._lazy_frames)

    def test_save(self, create_test_files: Path, tmp_path: Path) -> None:
        """Test save method raises NotImplementedError."""
        folder_path = create_test_files
        output_folder = tmp_path / "output"
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)

        # saveメソッドが未実装のため、NotImplementedErrorが発生することを確認
        with pytest.raises(
            NotImplementedError, match="The save method is not currently implemented."
        ):
            dataset.save(str(output_folder), filename_prefix="processed_")

        # SpectrogramFrameDatasetのsaveも同様に例外を発生させるか確認
        stft_dataset = dataset.stft()
        with pytest.raises(
            NotImplementedError, match="The save method is not currently implemented."
        ):
            stft_dataset.save(str(output_folder), filename_prefix="spec_")


# --- Test SpectrogramFrameDataset ---


class TestSpectrogramFrameDataset:
    def test_init_lazy(self, create_test_files: Path) -> None:
        """Test initialization (typically via ChannelFrameDataset.stft)."""
        folder_path = create_test_files
        channel_ds = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        stft_ds = channel_ds.stft()  # Creates SpectrogramFrameDataset

        assert isinstance(stft_ds, SpectrogramFrameDataset)
        assert stft_ds._lazy_loading is True
        assert stft_ds._source_dataset is channel_ds
        assert stft_ds._transform is not None
        assert len(stft_ds) == len(channel_ds)
        assert all(not lf.is_loaded for lf in stft_ds._lazy_frames)

    def test_load_file_not_implemented(self, tmp_path: Path) -> None:
        """Test that _load_file raises NotImplementedError."""
        # Create a dummy file that might look like a spectrogram source
        dummy_spec_file = tmp_path / "spec.npy"
        np.save(dummy_spec_file, np.random.rand(1, 10, 5))

        # Initialize directly (not the typical way)
        spec_ds = SpectrogramFrameDataset(str(tmp_path), file_extensions=[".npy"])

        # 現在の実装では、_load_fileは例外を投げるが、
        # _ensure_loadedでそれがキャッチされる
        # 直接_load_fileを呼び出して例外を確認
        with pytest.raises(
            NotImplementedError,
            match="No method defined for directly loading SpectrogramFrames",
        ):
            spec_ds._load_file(dummy_spec_file)

        # __getitem__経由では例外はキャッチされてNoneが返される
        result = spec_ds[0]
        assert result is None

    def test_apply(self, create_test_files: Path) -> None:
        """Test apply on SpectrogramFrameDataset."""
        folder_path = create_test_files
        channel_ds = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        stft_ds = channel_ds.stft()

        def spec_transform(spec_frame: SpectrogramFrame) -> SpectrogramFrame:
            # より単純な変換関数（amplitude_to_dbは実装されていない可能性があるため）
            return spec_frame + 1

        transformed_spec_ds = stft_ds.apply(spec_transform)

        assert isinstance(transformed_spec_ds, SpectrogramFrameDataset)
        assert transformed_spec_ds._lazy_loading is True
        assert transformed_spec_ds._source_dataset is stft_ds
        assert transformed_spec_ds._transform is spec_transform

        # Trigger computation
        original_spec_frame = stft_ds[0]  # Computes STFT
        transformed_spec_frame = transformed_spec_ds[0]  # Computes STFT then transform

        assert original_spec_frame is not None
        assert transformed_spec_frame is not None
        # Verify transformation (check if data changed)
        assert not np.allclose(
            original_spec_frame.compute(), transformed_spec_frame.compute()
        )

    @patch("matplotlib.pyplot.show")  # Prevent plots from displaying during tests
    def test_plot(
        self,
        mock_show: MagicMock,
        create_test_files: Path,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """Test plotting from SpectrogramFrameDataset."""
        folder_path = create_test_files
        channel_ds = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        stft_ds = channel_ds.stft()

        # Mock SpectrogramFrame.plot to check if it's called
        with patch.object(
            SpectrogramFrame, "plot", return_value=MagicMock(spec=Axes)
        ) as mock_frame_plot:
            stft_ds.plot(0)  # Plot the first spectrogram
            mock_frame_plot.assert_called_once()

        # Test plot with load/transform error
        caplog.clear()

        def failing_stft(frame: ChannelFrame) -> SpectrogramFrame:
            if frame.label == "test1":
                raise ValueError("STFT failed")
            return frame.stft()

        # Need to wrap this in a SpectrogramFrameDataset manually for plot method
        error_spec_ds = SpectrogramFrameDataset(
            folder_path=str(folder_path),
            source_dataset=channel_ds,
            transform=failing_stft,
        )

        with patch.object(SpectrogramFrame, "plot") as mock_plot_error:
            error_spec_ds.plot(0)  # Attempt to plot the failing one
            mock_plot_error.assert_not_called()
            assert any(
                "STFT failed" in rec.message
                for rec in caplog.records
                if rec.levelname == "WARNING"
            )

        # Test plot when frame has no plot method (simulate)
        caplog.clear()

        # plotがないクラスを定義
        with patch.object(SpectrogramFrame, "plot", None):
            # plotメソッドがないフレームをロードした場合のテスト
            stft_ds.plot(0)
            assert any(
                "Frame" in rec.message
                for rec in caplog.records
                if rec.levelname == "WARNING"
            )

        # Test error during plotting itself
        caplog.clear()
        plot_error_msg = "The save method is not currently implemented."
        with patch.object(
            SpectrogramFrame, "plot", side_effect=RuntimeError(plot_error_msg)
        ) as mock_plot_runtime_error:
            stft_ds.plot(1)  # Plot second item, assume it loads ok
            mock_plot_runtime_error.assert_called_once()
            assert any(
                plot_error_msg in rec.message
                for rec in caplog.records
                if rec.levelname == "ERROR"
            )


# --- Test _SampledFrameDataset (Internal Class) ---


class TestSampledFrameDataset:
    def test_init(self, create_test_files: Path) -> None:
        """Test initialization of _SampledFrameDataset."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)

        # サンプリングインデックスの準備
        sampled_indices = [0, 2]  # test1.wav と test3.csv を選択

        # _SampledFrameDataset の初期化
        sampled_ds = _SampledFrameDataset(dataset, sampled_indices)

        # 基本的なプロパティの確認
        assert sampled_ds._original_dataset is dataset
        assert sampled_ds._original_indices == sampled_indices
        assert len(sampled_ds) == len(sampled_indices)

        # LazyFrames の確認
        assert len(sampled_ds._lazy_frames) == len(sampled_indices)
        assert all(not lf.is_loaded for lf in sampled_ds._lazy_frames)
        assert sampled_ds._lazy_frames[0].file_path.name == "test1.wav"
        assert sampled_ds._lazy_frames[1].file_path.name == "test3.csv"

    def test_getitem(self, create_test_files: Path) -> None:
        """Test __getitem__ method of _SampledFrameDataset."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)

        # 元のデータセットからサンプリング
        sampled_indices = [0, 2]  # test1.wav, test3.csv
        sampled_ds = _SampledFrameDataset(dataset, sampled_indices)

        assert isinstance(sampled_ds, _SampledFrameDataset)
        assert sampled_ds._original_dataset is dataset
        assert sampled_ds._original_indices == sampled_indices

        # サンプリングされたデータセットからフレームを取得
        sampled_frame_0 = sampled_ds[0]

        # 元のデータセットから直接同じフレームを取得
        original_frame_0 = dataset[sampled_indices[0]]

        # Noneチェック
        assert sampled_frame_0 is not None
        assert original_frame_0 is not None

        # 同じフレームが取得できていることを確認
        assert isinstance(sampled_frame_0, ChannelFrame)
        assert sampled_frame_0.label == original_frame_0.label
        np.testing.assert_array_equal(
            sampled_frame_0.compute(),
            original_frame_0.compute(),
            err_msg="サンプリングデータセットのフレームが元のデータセットと一致しません",
        )

        # キャッシュが機能していることを確認
        assert sampled_ds._lazy_frames[0].is_loaded is True

        # キャッシュを通じて同じフレームオブジェクトが返されることを確認
        cached_frame = sampled_ds[0]
        assert cached_frame is sampled_frame_0  # 同じインスタンスが返されるはず

        # 2番目の要素も確認
        sampled_frame_1 = sampled_ds[1]
        original_frame_1 = dataset[sampled_indices[1]]

        assert isinstance(sampled_frame_1, ChannelFrame)
        assert sampled_frame_1 is not None
        assert original_frame_1 is not None
        assert sampled_frame_1.label == original_frame_1.label
        np.testing.assert_array_equal(
            sampled_frame_1.compute(), original_frame_1.compute()
        )

        # 範囲外のインデックスに対するエラーテスト
        with pytest.raises(IndexError):
            _ = sampled_ds[2]  # サンプリングデータセットは2つの要素しかない

        # 元のデータセットのロードエラーをシミュレート
        # _ensure_loadedメソッドを直接モックする方が確実です
        with patch.object(dataset, "_ensure_loaded", return_value=None):
            # _SampledFrameDatasetはキャッシュをリセットする必要がある
            sampled_ds._lazy_frames[0].reset()
            result = sampled_ds[0]
            assert result is None

    def test_apply(self, create_test_files: Path) -> None:
        """Test apply method of _SampledFrameDataset."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        sampled_ds = dataset.sample(n=1, seed=42)

        def transform_gain(f: ChannelFrame) -> ChannelFrame:
            assert f is not None, "Input frame to transform_gain is None"
            return f * 3

        transformed_sampled = sampled_ds.apply(transform_gain)

        # 基本的な検証
        assert isinstance(transformed_sampled, _SampledFrameDataset)
        assert transformed_sampled._original_dataset is not dataset
        assert transformed_sampled._original_dataset._transform is transform_gain
        assert transformed_sampled._original_dataset._source_dataset is dataset
        assert len(transformed_sampled) == 1
        assert all(not lf.is_loaded for lf in transformed_sampled._lazy_frames)

        # 元のフレームを取得
        original_index = sampled_ds._original_indices[0]  # type: ignore[attr-defined]
        original_frame = dataset[original_index]

        # サンプリングされたフレームを取得
        sampled_frame = sampled_ds[0]
        assert original_frame is not None
        assert sampled_frame is not None
        np.testing.assert_allclose(
            sampled_frame.compute(),
            original_frame.compute(),
            err_msg="Data mismatch between dataset and sampled_ds",
        )

        # 変換後のフレームを取得して検証
        final_frame = transformed_sampled[0]
        assert final_frame is not None
        assert original_frame is not None
        expected_data = original_frame.compute() * 3
        np.testing.assert_allclose(
            final_frame.compute(),
            expected_data,
            err_msg="Data mismatch after applying transform",
        )

        # キャッシュが機能していることを確認
        assert transformed_sampled._lazy_frames[0].is_loaded is True
        cached_final_frame = transformed_sampled[0]
        assert cached_final_frame is final_frame

    def test_load_file_not_implemented(self, create_test_files: Path) -> None:
        """Test that _load_file raises NotImplementedError."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)
        sampled_indices = [0, 1]
        sampled_ds = _SampledFrameDataset(dataset, sampled_indices)

        with pytest.raises(
            NotImplementedError,
            match="_SampledFrameDataset does not load files directly.",
        ):
            sampled_ds._load_file(Path("dummy.wav"))

    def test_sample_index_error(self, create_test_files: Path) -> None:
        """Test error handling with invalid indices."""
        folder_path = create_test_files
        dataset = ChannelFrameDataset(str(folder_path), lazy_loading=True)

        # 範囲外のインデックスを含むケース
        invalid_indices = [0, 10]  # 10は範囲外（ファイルは3つしかない）

        with pytest.raises(
            IndexError,
            match="Indices are out of range for the original dataset.",
        ):
            _SampledFrameDataset(dataset, invalid_indices)



import abc
import inspect
import logging
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Optional, TypeVar, Union

# Import librosa (including display)
import librosa

from wandas.utils.introspection import filter_kwargs

try:
    # Avoid error due to librosa.display not being explicitly exported
    from librosa import display  # type: ignore
except ImportError:
    # fallback
    display = librosa.display  # type: ignore

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.lines import Line2D

if TYPE_CHECKING:
    from wandas.core.base_frame import BaseFrame
    from wandas.frames.channel import ChannelFrame
    from wandas.frames.noct import NOctFrame
    from wandas.frames.spectral import SpectralFrame
    from wandas.frames.spectrogram import SpectrogramFrame

logger = logging.getLogger(__name__)

TFrame = TypeVar("TFrame", bound="BaseFrame[Any]")


class PlotStrategy(abc.ABC, Generic[TFrame]):
    """Base class for plotting strategies"""

    name: ClassVar[str]

    @abc.abstractmethod
    def channel_plot(self, x: Any, y: Any, ax: "Axes") -> None:
        """Implementation of channel plotting"""
        pass

    @abc.abstractmethod
    def plot(
        self,
        bf: TFrame,
        ax: Optional["Axes"] = None,
        title: Optional[str] = None,
        overlay: bool = False,
        **kwargs: Any,
    ) -> Union["Axes", Iterator["Axes"]]:
        """Implementation of plotting"""
        pass


# Helper function for return type
def _return_axes_iterator(axes_list: Any) -> Iterator[Axes]:
    """Helper to convert fig.axes to Iterator[Axes] with proper typing"""
    return iter(axes_list)


class WaveformPlotStrategy(PlotStrategy["ChannelFrame"]):
    """Strategy for waveform plotting"""

    name = "waveform"

    def channel_plot(
        self,
        x: Any,
        y: Any,
        ax: "Axes",
        **kwargs: Any,
    ) -> None:
        """Implementation of channel plotting"""
        ax.plot(x, y, **kwargs)
        ax.set_ylabel("Amplitude")
        ax.grid(True)
        if "label" in kwargs:
            ax.legend()

    def plot(
        self,
        bf: "ChannelFrame",
        ax: Optional["Axes"] = None,
        title: Optional[str] = None,
        overlay: bool = False,
        **kwargs: Any,
    ) -> Union["Axes", Iterator["Axes"]]:
        """Waveform plotting"""
        kwargs = kwargs or {}
        ylabel = kwargs.pop("ylabel", "Amplitude")
        xlabel = kwargs.pop("xlabel", "Time [s]")
        alpha = kwargs.pop("alpha", 1)
        plot_kwargs = filter_kwargs(
            Line2D,
            kwargs,
            strict_mode=True,
        )
        ax_set = filter_kwargs(
            Axes.set,
            kwargs,
            strict_mode=True,
        )
        if overlay:
            if ax is None:
                fig, ax = plt.subplots(figsize=(10, 4))
            self.channel_plot(
                bf.time, bf.data.T, ax, label=bf.labels, alpha=alpha, **plot_kwargs
            )
            ax.set(
                ylabel=ylabel,
                title=title or bf.label or "Channel Data",
                xlabel=xlabel,
                **ax_set,
            )
            if ax is None:
                fig.suptitle(title or bf.label or None)
                plt.tight_layout()
                plt.show()
            return ax
        else:
            num_channels = bf.n_channels
            fig, axs = plt.subplots(
                num_channels, 1, figsize=(10, 4 * num_channels), sharex=True
            )
            # Convert axs to list if it is a single Axes object
            if not isinstance(axs, (list, np.ndarray)):
                axs = [axs]

            axes_list = list(axs)
            data = bf.data
            for ax_i, channel_data, ch_meta in zip(axes_list, data, bf.channels):
                self.channel_plot(
                    bf.time, channel_data, ax_i, alpha=alpha, **plot_kwargs
                )
                ax_i.set(
                    ylabel=ylabel,
                    title=ch_meta.label,
                    **ax_set,
                )

            axes_list[-1].set(
                ylabel=ylabel,
                xlabel="Time [s]",
            )
            fig.suptitle(title or bf.label or "Channel Data")

            if ax is None:
                plt.tight_layout()
                plt.show()

            return _return_axes_iterator(fig.axes)


class FrequencyPlotStrategy(PlotStrategy["SpectralFrame"]):
    """Strategy for frequency domain plotting"""

    name = "frequency"

    def channel_plot(
        self,
        x: Any,
        y: Any,
        ax: "Axes",
        **kwargs: Any,
    ) -> None:
        """Implementation of channel plotting"""
        ax.plot(x, y, **kwargs)
        ax.grid(True)
        if "label" in kwargs:
            ax.legend()

    def plot(
        self,
        bf: "SpectralFrame",
        ax: Optional["Axes"] = None,
        title: Optional[str] = None,
        overlay: bool = False,
        **kwargs: Any,
    ) -> Union["Axes", Iterator["Axes"]]:
        """Frequency domain plotting"""
        kwargs = kwargs or {}
        is_aw = kwargs.pop("Aw", False)
        if bf.operation_history[-1]["operation"] == "coherence":
            unit = ""
            data = bf.magnitude
            ylabel = kwargs.pop("ylabel", "coherence")
        else:
            if is_aw:
                unit = "dBA"
                data = bf.dBA
            else:
                unit = "dB"
                data = bf.dB
            ylabel = kwargs.pop("ylabel", f"Spectrum level [{unit}]")
        xlabel = kwargs.pop("xlabel", "Frequency [Hz]")
        alpha = kwargs.pop("alpha", 1)
        plot_kwargs = filter_kwargs(Line2D, kwargs, strict_mode=True)
        ax_set = filter_kwargs(Axes.set, kwargs, strict_mode=True)
        if overlay:
            if ax is None:
                _, ax = plt.subplots(figsize=(10, 4))
            self.channel_plot(
                bf.freqs,
                data.T,
                ax,
                label=bf.labels,
                alpha=alpha,
                **plot_kwargs,
            )
            ax.set(
                ylabel=ylabel,
                xlabel=xlabel,
                title=title or bf.label or "Channel Data",
                **ax_set,
            )
            if ax is None:
                plt.tight_layout()
                plt.show()
            return ax
        else:
            num_channels = bf.n_channels
            fig, axs = plt.subplots(
                num_channels, 1, figsize=(10, 4 * num_channels), sharex=True
            )
            # Convert axs to list if it is a single Axes object
            if not isinstance(axs, (list, np.ndarray)):
                axs = [axs]

            axes_list = list(axs)
            for ax_i, channel_data, ch_meta in zip(axes_list, data, bf.channels):
                self.channel_plot(
                    bf.freqs,
                    channel_data,
                    ax_i,
                    label=ch_meta.label,
                    alpha=alpha,
                    **plot_kwargs,
                )
                ax_i.set(
                    ylabel=ylabel,
                    title=ch_meta.label,
                    xlabel=xlabel,
                    **ax_set,
                )
            axes_list[-1].set(ylabel=ylabel, xlabel=xlabel)
            fig.suptitle(title or bf.label or "Channel Data")
            if ax is None:
                plt.tight_layout()
                plt.show()
            return _return_axes_iterator(fig.axes)


class NOctPlotStrategy(PlotStrategy["NOctFrame"]):
    """Strategy for N-octave band analysis plotting"""

    name = "noct"

    def channel_plot(
        self,
        x: Any,
        y: Any,
        ax: "Axes",
        **kwargs: Any,
    ) -> None:
        """Implementation of channel plotting"""
        ax.step(x, y, **kwargs)
        ax.grid(True)
        if "label" in kwargs:
            ax.legend()

    def plot(
        self,
        bf: "NOctFrame",
        ax: Optional["Axes"] = None,
        title: Optional[str] = None,
        overlay: bool = False,
        **kwargs: Any,
    ) -> Union["Axes", Iterator["Axes"]]:
        """N-octave band analysis plotting"""
        kwargs = kwargs or {}
        is_aw = kwargs.pop("Aw", False)

        if is_aw:
            unit = "dBrA"
            data = bf.dBA
        else:
            unit = "dBr"
            data = bf.dB
        ylabel = kwargs.pop("ylabel", f"Spectrum level [{unit}]")
        xlabel = kwargs.pop("xlabel", "Center frequency [Hz]")
        alpha = kwargs.pop("alpha", 1)
        plot_kwargs = filter_kwargs(Line2D, kwargs, strict_mode=True)
        ax_set = filter_kwargs(Axes.set, kwargs, strict_mode=True)
        if overlay:
            if ax is None:
                _, ax = plt.subplots(figsize=(10, 4))
            self.channel_plot(
                bf.freqs,
                data.T,
                ax,
                label=bf.labels,
                alpha=alpha,
                **plot_kwargs,
            )
            ax.set(
                ylabel=ylabel,
                xlabel=xlabel,
                title=title or bf.label or f"1/{str(bf.n)}-Octave Spectrum",
                **ax_set,
            )
            if ax is None:
                plt.tight_layout()
                plt.show()
            return ax
        else:
            num_channels = bf.n_channels
            fig, axs = plt.subplots(
                num_channels, 1, figsize=(10, 4 * num_channels), sharex=True
            )
            # Convert axs to list if it is a single Axes object
            if not isinstance(axs, (list, np.ndarray)):
                axs = [axs]

            axes_list = list(axs)
            for ax_i, channel_data, ch_meta in zip(axes_list, data, bf.channels):
                self.channel_plot(
                    bf.freqs,
                    channel_data,
                    ax_i,
                    label=ch_meta.label,
                    alpha=alpha,
                    **plot_kwargs,
                )
                ax_i.set(
                    ylabel=ylabel,
                    title=ch_meta.label,
                    xlabel=xlabel,
                    **ax_set,
                )
            axes_list[-1].set(ylabel=ylabel, xlabel=xlabel)
            fig.suptitle(title or bf.label or f"1/{str(bf.n)}-Octave Spectrum")
            if ax is None:
                plt.tight_layout()
                plt.show()
            return _return_axes_iterator(fig.axes)


class SpectrogramPlotStrategy(PlotStrategy["SpectrogramFrame"]):
    """Strategy for spectrogram plotting"""

    name = "spectrogram"

    def channel_plot(
        self,
        x: Any,
        y: Any,
        ax: "Axes",
        **kwargs: Any,
    ) -> None:
        """Implementation of channel plotting"""
        pass

    def plot(
        self,
        bf: "SpectrogramFrame",
        ax: Optional["Axes"] = None,
        title: Optional[str] = None,
        overlay: bool = False,
        **kwargs: Any,
    ) -> Union["Axes", Iterator["Axes"]]:
        """Spectrogram plotting"""
        if overlay:
            raise ValueError("Overlay is not supported for SpectrogramPlotStrategy.")

        if ax is not None and bf.n_channels > 1:
            raise ValueError("ax must be None when n_channels > 1.")

        kwargs = kwargs or {}

        is_aw = kwargs.pop("Aw", False)
        if is_aw:
            unit = "dBA"
            data = bf.dBA
        else:
            unit = "dB"
            data = bf.dB

        specshow_kwargs = filter_kwargs(display.specshow, kwargs, strict_mode=True)
        ax_set_kwargs = filter_kwargs(Axes.set, kwargs, strict_mode=True)

        cmap = kwargs.pop("cmap", "jet")
        vmin = kwargs.pop("vmin", None)
        vmax = kwargs.pop("vmax", None)

        if ax is not None:
            img = display.specshow(
                data=data[0],
                sr=bf.sampling_rate,
                hop_length=bf.hop_length,
                n_fft=bf.n_fft,
                win_length=bf.win_length,
                x_axis="time",
                y_axis="linear",
                cmap=cmap,
                ax=ax,
                vmin=vmin,
                vmax=vmax,
                **specshow_kwargs,
            )
            ax.set(
                title=title or bf.label or "Spectrogram",
                ylabel="Frequency [Hz]",
                xlabel="Time [s]",
                **ax_set_kwargs,
            )

            fig = ax.figure
            if fig is not None:
                cbar = fig.colorbar(img, ax=ax)
                cbar.set_label(f"Spectrum level [{unit}]")
            return ax

        else:
            # Create a new figure if ax is None
            num_channels = bf.n_channels
            fig, axs = plt.subplots(
                num_channels, 1, figsize=(10, 5 * num_channels), sharex=True
            )
            if not isinstance(fig, plt.Figure):
                raise ValueError("fig must be a matplotlib Figure object.")
            # Convert axs to array if it is a single Axes object
            if not isinstance(axs, np.ndarray):
                axs = np.array([axs])

            for ax_i, channel_data, ch_meta in zip(axs.flatten(), data, bf.channels):
                img = display.specshow(
                    data=channel_data,
                    sr=bf.sampling_rate,
                    hop_length=bf.hop_length,
                    n_fft=bf.n_fft,
                    win_length=bf.win_length,
                    x_axis="time",
                    y_axis="linear",
                    ax=ax_i,
                    cmap=cmap,
                    vmin=vmin,
                    vmax=vmax,
                    **specshow_kwargs,
                )
                ax_i.set(
                    title=ch_meta.label,
                    ylabel="Frequency [Hz]",
                    xlabel="Time [s]",
                    **ax_set_kwargs,
                )
                cbar = ax_i.figure.colorbar(img, ax=ax_i)
                cbar.set_label(f"Spectrum level [{unit}]")
                fig.suptitle(title or "Spectrogram Data")
            plt.tight_layout()
            plt.show()

            return _return_axes_iterator(fig.axes)


class DescribePlotStrategy(PlotStrategy["ChannelFrame"]):
    """Strategy for visualizing ChannelFrame data with describe plot"""

    name = "describe"

    def channel_plot(self, x: Any, y: Any, ax: "Axes", **kwargs: Any) -> None:
        """Implementation of channel plotting"""
        pass  # This method is not used for describe plot

    def plot(
        self,
        bf: "ChannelFrame",
        ax: Optional["Axes"] = None,
        title: Optional[str] = None,
        overlay: bool = False,
        **kwargs: Any,
    ) -> Union["Axes", Iterator["Axes"]]:
        """Implementation of describe method for visualizing ChannelFrame data"""

        fmin = kwargs.pop("fmin", 0)
        fmax = kwargs.pop("fmax", None)
        cmap = kwargs.pop("cmap", "jet")
        vmin = kwargs.pop("vmin", None)
        vmax = kwargs.pop("vmax", None)
        xlim = kwargs.pop("xlim", None)
        ylim = kwargs.pop("ylim", None)
        is_aw = kwargs.pop("Aw", False)
        waveform = kwargs.pop("waveform", {})
        spectral = kwargs.pop("spectral", dict(xlim=(vmin, vmax)))

        gs = gridspec.GridSpec(2, 3, height_ratios=[1, 3], width_ratios=[3, 1, 0.1])
        gs.update(wspace=0.2)

        fig = plt.figure(figsize=(12, 6))
        fig.subplots_adjust(wspace=0.0001)

        # First subplot (Time Plot)
        ax_1 = fig.add_subplot(gs[0])
        bf.plot(plot_type="waveform", ax=ax_1, overlay=True)
        ax_1.set(**waveform)
        ax_1.legend().set_visible(False)
        ax_1.set(xlabel="", title="")

        # Second subplot (STFT Plot)
        ax_2 = fig.add_subplot(gs[3], sharex=ax_1)
        stft_ch = bf.stft()
        if is_aw:
            unit = "dBA"
            channel_data = stft_ch.dBA[0]
        else:
            unit = "dB"
            channel_data = stft_ch.dB[0]
        # Get the maximum value of the data and round it to a convenient value
        if vmax is None:
            data_max = np.nanmax(channel_data)
            # Round to a convenient number with increments of 10, 5, or 2
            for step in [10, 5, 2]:
                rounded_max = np.ceil(data_max / step) * step
                if rounded_max >= data_max:
                    vmax = rounded_max
                    vmin = vmax - 180
                    break
        img = display.specshow(
            data=channel_data,
            sr=bf.sampling_rate,
            hop_length=stft_ch.hop_length,
            n_fft=stft_ch.n_fft,
            win_length=stft_ch.win_length,
            x_axis="time",
            y_axis="linear",
            ax=ax_2,
            fmin=fmin,
            fmax=fmax,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )
        ax_2.set(xlim=xlim, ylim=ylim)

        # Third subplot
        ax_3 = fig.add_subplot(gs[1])
        ax_3.axis("off")

        # Fourth subplot (Welch Plot)
        ax_4 = fig.add_subplot(gs[4], sharey=ax_2)
        welch_ch = bf.welch()
        if is_aw:
            unit = "dBA"
            data_db = welch_ch.dBA
        else:
            unit = "dB"
            data_db = welch_ch.dB
        ax_4.plot(data_db.T, welch_ch.freqs.T)
        ax_4.grid(True)
        ax_4.set(xlabel=f"Spectrum level [{unit}]", **spectral)

        cbar = fig.colorbar(img, ax=ax_4, format="%+2.0f")
        cbar.set_label(unit)
        fig.suptitle(title or bf.label or "Channel Data")

        return _return_axes_iterator(fig.axes)


class MatrixPlotStrategy(PlotStrategy[Union["SpectralFrame"]]):
    """Strategy for displaying relationships between channels in matrix format"""

    name = "matrix"

    def channel_plot(
        self,
        x: Any,
        y: Any,
        ax: "Axes",
        title: Optional[str] = None,
        ylabel: str = "",
        xlabel: str = "Frequency [Hz]",
        alpha: float = 0,
        **kwargs: Any,
    ) -> None:
        ax.plot(x, y, **kwargs)
        ax.grid(True)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(title or "")

    def plot(
        self,
        bf: "SpectralFrame",
        ax: Optional["Axes"] = None,
        title: Optional[str] = None,
        overlay: bool = False,
        **kwargs: Any,
    ) -> Union["Axes", Iterator["Axes"]]:
        kwargs = kwargs or {}
        is_aw = kwargs.pop("Aw", False)
        if (
            len(bf.operation_history) > 0
            and bf.operation_history[-1]["operation"] == "coherence"
        ):
            unit = ""
            data = bf.magnitude
            ylabel = kwargs.pop("ylabel", "coherence")
        else:
            if is_aw:
                unit = "dBA"
                data = bf.dBA
            else:
                unit = "dB"
                data = bf.dB
            ylabel = kwargs.pop("ylabel", f"Spectrum level [{unit}]")
        xlabel = kwargs.pop("xlabel", "Frequency [Hz]")
        alpha = kwargs.pop("alpha", 1)
        plot_kwargs = filter_kwargs(Line2D, kwargs, strict_mode=True)
        ax_set = filter_kwargs(Axes.set, kwargs, strict_mode=True)
        num_channels = bf.n_channels
        if overlay:
            if ax is None:
                fig, ax = plt.subplots(1, 1, figsize=(6, 6))
            self.channel_plot(
                bf.freqs,
                data.T,
                ax,  # ここで必ずAxes型
                title=title or bf.label or "Spectral Data",
                ylabel=ylabel,
                xlabel=xlabel,
                alpha=alpha,
                **plot_kwargs,
            )
            if ax is not None:
                ax.set(**ax_set)
                fig.suptitle(title or bf.label or "Spectral Data")
                plt.tight_layout()
                plt.show()
                return ax
            # axがNoneのケースはここで発生しない
        else:
            num_rows = int(np.ceil(np.sqrt(num_channels)))
            fig, axs = plt.subplots(
                num_rows,
                num_rows,
                figsize=(3 * num_rows, 3 * num_rows),
                sharex=True,
                sharey=True,
            )
            if isinstance(axs, np.ndarray):
                axes_list = axs.flatten().tolist()
            elif isinstance(axs, list):
                import itertools

                axes_list = list(itertools.chain.from_iterable(axs))
            else:
                axes_list = [axs]
            for ax_i, channel_data, ch_meta in zip(axes_list, data, bf.channels):
                self.channel_plot(
                    bf.freqs,
                    channel_data,
                    ax_i,
                    title=ch_meta.label,
                    ylabel=ylabel,
                    xlabel=xlabel,
                    alpha=alpha,
                    **plot_kwargs,
                )
                ax_i.set(**ax_set)
            fig.suptitle(title or bf.label or "Spectral Data")
            plt.tight_layout()
            plt.show()
            return _return_axes_iterator(fig.axes)

        raise NotImplementedError()


# Maintain mapping of plot types to corresponding classes
_plot_strategies: dict[str, type[PlotStrategy[Any]]] = {}


def register_plot_strategy(strategy_cls: type) -> None:
    """Register a new plot strategy from a class"""
    if not issubclass(strategy_cls, PlotStrategy):
        raise TypeError("Strategy class must inherit from PlotStrategy.")
    if inspect.isabstract(strategy_cls):
        raise TypeError("Cannot register abstract PlotStrategy class.")
    _plot_strategies[strategy_cls.name] = strategy_cls


# Modified to auto-register only non-abstract subclasses
for strategy_cls in PlotStrategy.__subclasses__():
    if not inspect.isabstract(strategy_cls):
        register_plot_strategy(strategy_cls)


def get_plot_strategy(name: str) -> type[PlotStrategy[Any]]:
    """Get plot strategy by name"""
    if name not in _plot_strategies:
        raise ValueError(f"Unknown plot type: {name}")
    return _plot_strategies[name]


def create_operation(name: str, **params: Any) -> PlotStrategy[Any]:
    """Create operation instance from operation name and parameters"""
    operation_class = get_plot_strategy(name)
    return operation_class(**params)



from unittest import mock

import dask.array as da
import numpy as np
from dask.array.core import Array as DaArray

from wandas.frames.channel import ChannelFrame
from wandas.frames.noct import NOctFrame
from wandas.frames.spectral import SpectralFrame
from wandas.frames.spectrogram import SpectrogramFrame

_da_from_array = da.from_array  # type: ignore [unused-ignore]


class TestChannelTransform:
    def setup_method(self) -> None:
        """Set up test fixtures for each test."""
        # Create a simple dask array for testing
        self.sample_rate: float = 16000
        self.data: np.ndarray = np.random.random((2, 16000))  # 2 channels, 1 second
        self.dask_data: DaArray = _da_from_array(self.data, chunks=(1, 4000))
        self.channel_frame: ChannelFrame = ChannelFrame(
            data=self.dask_data, sampling_rate=self.sample_rate, label="test_audio"
        )

    def test_fft_transform(self) -> None:
        """Test fft method for lazy transformation to frequency domain."""
        from wandas.processing import FFT

        with mock.patch("wandas.processing.create_operation") as mock_create_op:
            # モックFFTオペレーションの設定
            mock_fft = mock.MagicMock(spec=FFT)
            mock_fft.n_fft = 4096
            mock_fft.window = "hann"
            mock_data = mock.MagicMock(spec=DaArray)
            mock_data.ndim = 2  # Set ndim property to pass dimension check
            mock_data.shape = (2, 2049)  # Set appropriate shape for a 2D array
            mock_fft.process.return_value = mock_data
            mock_create_op.return_value = mock_fft

            # fftを遅延実行
            result = self.channel_frame.fft(n_fft=4096, window="hamming")

            # オペレーションが正しく作成されたか確認
            mock_create_op.assert_called_with(
                "fft", self.sample_rate, n_fft=4096, window="hamming"
            )

            # processメソッドが呼び出されたか確認
            mock_fft.process.assert_called_once_with(self.channel_frame._data)

            # 結果が正しい型か確認
            assert isinstance(result, SpectralFrame)
            assert result.n_fft == 4096
            assert result.window == "hann"
            assert result.previous is self.channel_frame

    def test_welch_transform(self) -> None:
        """
        Test welch method for lazy transformation to frequency domain
        using Welch's method.
        """
        from wandas.processing import Welch

        with mock.patch("wandas.processing.create_operation") as mock_create_op:
            # モックWelchオペレーションの設定
            mock_welch = mock.MagicMock(spec=Welch)
            mock_welch.n_fft = 2048
            mock_welch.hop_length = 256
            mock_welch.win_length = 1024
            mock_welch.window = "blackman"
            mock_welch.average = "mean"
            mock_data = mock.MagicMock(spec=DaArray)
            mock_data.ndim = 2  # Set ndim property to pass dimension check
            mock_data.shape = (2, 1025)  # Set appropriate shape for a 2D array
            mock_welch.process.return_value = mock_data
            mock_create_op.return_value = mock_welch

            # welchを遅延実行
            result = self.channel_frame.welch(
                n_fft=2048,
                hop_length=256,
                win_length=1024,
                window="blackman",
                average="mean",
            )

            # オペレーションが正しく作成されたか確認
            mock_create_op.assert_called_with(
                "welch",
                self.sample_rate,
                n_fft=2048,
                hop_length=256,
                win_length=1024,
                window="blackman",
                average="mean",
            )

            # processメソッドが呼び出されたか確認
            mock_welch.process.assert_called_once_with(self.channel_frame._data)

            # 結果が正しい型か確認
            assert isinstance(result, SpectralFrame)
            assert result.n_fft == 2048
            assert result.window == "blackman"
            assert result.previous is self.channel_frame

    def test_stft_transform(self) -> None:
        """Test stft method for lazy short-time Fourier transform."""
        from wandas.processing import STFT

        with mock.patch("wandas.processing.create_operation") as mock_create_op:
            # モックSTFTオペレーションの設定
            mock_stft = mock.MagicMock(spec=STFT)
            mock_data = mock.MagicMock(spec=DaArray)
            mock_data.ndim = 3  # Set ndim property to pass dimension check
            mock_data.shape = (2, 1025, 10)  # Set appropriate shape for a 3D array
            mock_stft.process.return_value = mock_data
            mock_create_op.return_value = mock_stft

            # stftを遅延実行（デフォルト引数）
            result = self.channel_frame.stft()

            # デフォルトパラメータの確認
            mock_create_op.assert_called_with(
                "stft",
                self.sample_rate,
                n_fft=2048,
                hop_length=512,  # n_fft//4
                win_length=2048,
                window="hann",
            )

            # processメソッドが呼び出されたか確認
            mock_stft.process.assert_called_once_with(self.channel_frame._data)

            # 結果が正しい型か確認
            assert isinstance(result, SpectrogramFrame)
            assert result.n_fft == 2048
            assert result.hop_length == 512
            assert result.win_length == 2048
            assert result.window == "hann"

            # カスタムパラメータでテスト
            mock_create_op.reset_mock()
            mock_stft.process.reset_mock()

            # Update mock data shape for n_fft=1024
            mock_data.shape = (
                2,
                513,
                10,
            )  # For n_fft=1024, freq_bins = 1024 // 2 + 1 = 513

            result = self.channel_frame.stft(
                n_fft=1024,
                hop_length=256,
                win_length=1024,
                window="hamming",
            )

            mock_create_op.assert_called_with(
                "stft",
                self.sample_rate,
                n_fft=1024,
                hop_length=256,
                win_length=1024,
                window="hamming",
            )

            assert result.n_fft == 1024
            assert result.hop_length == 256
            assert result.win_length == 1024
            assert result.window == "hamming"

    def test_noct_spectrum_transform(self) -> None:
        """Test noct_spectrum method for calculating N-octave spectrum analysis."""
        from wandas.processing import NOctSpectrum

        with mock.patch("wandas.processing.create_operation") as mock_create_op:
            # モックNOctSpectrumオペレーションの設定
            mock_noct = mock.MagicMock(spec=NOctSpectrum)
            mock_data = mock.MagicMock(spec=DaArray)
            mock_data.ndim = 2
            mock_data.shape = (2, 10)  # バンド数に応じた適切な形状
            mock_noct.process.return_value = mock_data
            mock_create_op.return_value = mock_noct

            # noct_spectrumを呼び出す
            fmin, fmax, n = 20, 20000, 3
            G, fr = 10, 1000  # noqa: N806
            result = self.channel_frame.noct_spectrum(
                fmin=fmin, fmax=fmax, n=n, G=G, fr=fr
            )

            # オペレーションが正しく作成されたか確認
            mock_create_op.assert_called_with(
                "noct_spectrum",
                self.sample_rate,
                fmin=fmin,
                fmax=fmax,
                n=n,
                G=G,
                fr=fr,
            )

            # processメソッドが呼び出されたか確認
            mock_noct.process.assert_called_once_with(self.channel_frame._data)

            # 結果が正しい型か確認
            assert isinstance(result, NOctFrame)
            assert result.fmin == fmin
            assert result.fmax == fmax
            assert result.n == n
            assert result.G == G
            assert result.fr == fr
            assert result.previous is self.channel_frame

    def test_csd(self) -> None:
        """クロススペクトル密度（CSD）メソッドのテスト"""
        # テスト用信号を作成
        sr = 1000
        t = np.linspace(0, 1, sr, endpoint=False)

        # 既知の周波数成分を持つ信号を作成（100Hzと200Hz）
        sig1 = np.sin(2 * np.pi * 100 * t) + 0.5 * np.sin(2 * np.pi * 200 * t)
        sig2 = 0.8 * np.sin(2 * np.pi * 100 * t) + 0.3 * np.sin(2 * np.pi * 300 * t)

        # 2チャンネル信号を作成
        sigs = np.array([sig1, sig2])

        # ChannelFrameインスタンスを作成
        cf = ChannelFrame.from_numpy(sigs, sr)

        # CSD計算のパラメータ
        n_fft = 512
        win_length = 256
        hop_length = 128

        # ChannelFrameメソッドを使用してCSDを計算
        csd_frame = cf.csd(
            n_fft=n_fft, win_length=win_length, hop_length=hop_length, window="hamming"
        )

        # 実際のデータを取得するために計算
        csd_data = csd_frame.compute()

        # 形状を確認
        assert csd_data.shape == (4, n_fft // 2 + 1)

        # チャンネルラベルが正しいことを確認
        ch_pairs = [ch.label for ch in csd_frame._channel_metadata]

        # 正しいラベル名を取得
        ch0_label = cf.channels[0].label
        ch1_label = cf.channels[1].label

        expected_pairs = [
            f"csd({ch0_label}, {ch0_label})",
            f"csd({ch0_label}, {ch1_label})",
            f"csd({ch1_label}, {ch0_label})",
            f"csd({ch1_label}, {ch1_label})",
        ]

        for pair in expected_pairs:
            assert pair in ch_pairs

        # 主要な周波数成分の存在を確認
        freq_bins = np.fft.rfftfreq(n_fft, d=1 / sr)
        idx_100hz = np.argmin(np.abs(freq_bins - 100))
        idx_200hz = np.argmin(np.abs(freq_bins - 200))
        idx_300hz = np.argmin(np.abs(freq_bins - 300))

        # 自己クロススペクトル密度では対応する周波数でピークが見られるはず
        # ch0 (sig1) の自己CSDは100Hzと200Hzでピークを持つ
        ch0_auto = csd_data[0]  # ch0 -> ch0
        assert abs(ch0_auto[idx_100hz]) > abs(ch0_auto[idx_300hz])
        assert abs(ch0_auto[idx_200hz]) > abs(ch0_auto[idx_300hz])

        # ch1 (sig2) の自己CSDは100Hzと300Hzでピークを持つ
        ch1_auto = csd_data[3]  # ch1 -> ch1
        assert abs(ch1_auto[idx_100hz]) > abs(ch1_auto[idx_200hz])
        assert abs(ch1_auto[idx_300hz]) > abs(ch1_auto[idx_200hz])

    def test_transfer_function(self) -> None:
        """伝達関数メソッドのテスト"""
        # 単純なシステム用のテスト信号を作成
        sr = 1000
        t = np.linspace(0, 1, sr, endpoint=False)

        # 入力信号
        input_sig = np.sin(2 * np.pi * 100 * t) + 0.5 * np.sin(2 * np.pi * 200 * t)

        # 出力信号（100Hzでゲイン2.0、200Hzでゲイン1.5を持つ）
        # より現実的にするためにノイズを追加
        output_sig = (
            2 * np.sin(2 * np.pi * 100 * t)
            + 0.75 * np.sin(2 * np.pi * 200 * t)
            + 0.05 * np.random.randn(len(t))
        )

        # 2チャンネル信号を作成
        sigs = np.array([input_sig, output_sig])

        # ChannelFrameインスタンスを作成
        cf = ChannelFrame.from_numpy(sigs, sr)

        # 伝達関数計算のパラメータ
        n_fft = 512
        win_length = 256
        hop_length = 128

        # ChannelFrameメソッドを使用して伝達関数を計算
        tf_frame = cf.transfer_function(
            n_fft=n_fft, win_length=win_length, hop_length=hop_length, window="hamming"
        )

        # 実際のデータを取得するために計算
        tf_data = tf_frame.compute()

        # 形状を確認
        assert tf_data.shape == (4, n_fft // 2 + 1)

        # テスト周波数のインデックスを探す
        freq_bins = np.fft.rfftfreq(n_fft, 1 / sr)
        idx_100hz = np.argmin(np.abs(freq_bins - 100))
        idx_200hz = np.argmin(np.abs(freq_bins - 200))

        # 入力から出力への伝達関数（インデックス1は入力->出力）
        # 主要周波数でのゲインがほぼ正確かチェック
        h_in_to_out = tf_data[1]
        assert np.isclose(np.abs(h_in_to_out[idx_100hz]), 2.0, rtol=0.2)
        assert np.isclose(np.abs(h_in_to_out[idx_200hz]), 1.5, rtol=0.2)

        # 自己伝達関数は約1.0になるはず
        assert np.isclose(np.abs(tf_data[0, idx_100hz]), 1.0, rtol=0.2)  # 入力->入力
        assert np.isclose(np.abs(tf_data[3, idx_100hz]), 1.0, rtol=0.2)  # 出力->出力

        # チャンネルラベルが正しいことを確認
        ch_pairs = [ch.label for ch in tf_frame._channel_metadata]

        # 正しいラベル名を取得
        ch0_label = cf.channels[0].label
        ch1_label = cf.channels[1].label

        expected_pairs = [
            f"$H_{{{ch0_label}, {ch0_label}}}$",
            f"$H_{{{ch0_label}, {ch1_label}}}$",
            f"$H_{{{ch1_label}, {ch0_label}}}$",
            f"$H_{{{ch1_label}, {ch1_label}}}$",
        ]

        for pair in expected_pairs:
            assert pair in ch_pairs

    def test_coherence(self) -> None:
        """コヒーレンスメソッドのテスト"""
        # テスト用信号を作成
        sr = 1000
        t = np.linspace(0, 1, sr, endpoint=False)

        # 関連性のある2つの信号を作成
        sig1 = np.sin(2 * np.pi * 100 * t) + 0.5 * np.sin(2 * np.pi * 200 * t)
        sig2 = (
            0.7 * np.sin(2 * np.pi * 100 * t)
            + 0.4 * np.sin(2 * np.pi * 200 * t)
            + 0.3 * np.sin(2 * np.pi * 300 * t)
        )

        # 2チャンネル信号を作成
        sigs = np.array([sig1, sig2])

        # ChannelFrameインスタンスを作成
        cf = ChannelFrame.from_numpy(sigs, sr)

        # コヒーレンス計算のパラメータ
        n_fft = 512
        win_length = 256
        hop_length = 128

        # ChannelFrameメソッドを使用してコヒーレンスを計算
        coherence_frame = cf.coherence(
            n_fft=n_fft, win_length=win_length, hop_length=hop_length, window="hamming"
        )

        # 実際のデータを取得するために計算
        coherence_data = coherence_frame.compute()

        # 形状を確認
        assert coherence_data.shape == (4, n_fft // 2 + 1)

        # 主要な周波数bins
        freq_bins = np.fft.rfftfreq(n_fft, d=1 / sr)
        idx_100hz = np.argmin(np.abs(freq_bins - 100))
        idx_300hz = np.argmin(np.abs(freq_bins - 300))

        # コヒーレンスの範囲チェック（0～1）
        # 数値計算の誤差を考慮して、わずかに1を超える値も許容する
        assert np.all(coherence_data >= 0)
        assert np.all(coherence_data <= 1.001)  # 許容誤差を追加

        # 自己コヒーレンスは1.0に近いはず
        assert np.isclose(
            coherence_data[0, idx_100hz], 1.0, rtol=1e-5
        )  # チャンネル0自己コヒーレンス
        assert np.isclose(
            coherence_data[3, idx_100hz], 1.0, rtol=1e-5
        )  # チャンネル1自己コヒーレンス

        # チャンネル間のコヒーレンスを100Hzと300Hzで確認
        # 100Hzでは両方の信号に成分があるので高いコヒーレンス
        # 300Hzではチャンネル2にのみ成分があるので低いコヒーレンス
        assert coherence_data[1, idx_100hz] > 0.8  # 100Hzでの高いコヒーレンス
        assert coherence_data[1, idx_300hz] < 0.5  # 300Hzでの低いコヒーレンス

        # チャンネルラベルが正しいことを確認
        ch_pairs = [ch.label for ch in coherence_frame._channel_metadata]

        # 正しいラベル名を取得
        ch0_label = cf.channels[0].label
        ch1_label = cf.channels[1].label

        expected_pairs = [
            f"$\\gamma_{{{ch0_label}, {ch0_label}}}$",
            f"$\\gamma_{{{ch0_label}, {ch1_label}}}$",
            f"$\\gamma_{{{ch1_label}, {ch0_label}}}$",
            f"$\\gamma_{{{ch1_label}, {ch1_label}}}$",
        ]

        for pair in expected_pairs:
            assert pair in ch_pairs



import logging
import random
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Generic, Optional, TypeVar, Union, cast, overload

from tqdm.auto import tqdm

from wandas.frames.channel import ChannelFrame
from wandas.frames.spectrogram import SpectrogramFrame

logger = logging.getLogger(__name__)

FrameType = Union[ChannelFrame, SpectrogramFrame]
F = TypeVar("F", bound=FrameType)
F_out = TypeVar("F_out", bound=FrameType)


@dataclass
class LazyFrame(Generic[F]):
    """
    A class that encapsulates a frame and its loading state.

    Attributes:
        file_path: File path associated with the frame
        frame: Loaded frame object (None if not loaded)
        is_loaded: Flag indicating if the frame is loaded
        load_attempted: Flag indicating if loading was attempted (for error detection)
    """

    file_path: Path
    frame: Optional[F] = None
    is_loaded: bool = False
    load_attempted: bool = False

    def ensure_loaded(self, loader: Callable[[Path], Optional[F]]) -> Optional[F]:
        """
        Ensures the frame is loaded, loading it if necessary.

        Args:
            loader: Function to load a frame from a file path

        Returns:
            The loaded frame, or None if loading failed
        """
        # Return the current frame if already loaded
        if self.is_loaded:
            return self.frame

        # Attempt to load if not loaded yet
        try:
            self.load_attempted = True
            self.frame = loader(self.file_path)
            self.is_loaded = True
            return self.frame
        except Exception as e:
            logger.error(f"Failed to load file {self.file_path}: {str(e)}")
            self.is_loaded = True  # Loading was attempted
            self.frame = None
            return None

    def reset(self) -> None:
        """
        Reset the frame state.
        """
        self.frame = None
        self.is_loaded = False
        self.load_attempted = False


class FrameDataset(Generic[F], ABC):
    """
    Abstract base dataset class for processing files in a folder.
    Includes lazy loading capability to efficiently handle large datasets.
    Subclasses handle specific frame types (ChannelFrame, SpectrogramFrame, etc.).
    """

    def __init__(
        self,
        folder_path: str,
        sampling_rate: Optional[int] = None,
        signal_length: Optional[int] = None,
        file_extensions: Optional[list[str]] = None,
        lazy_loading: bool = True,
        recursive: bool = False,
        source_dataset: Optional["FrameDataset[Any]"] = None,
        transform: Optional[Callable[[Any], Optional[F]]] = None,
    ):
        self.folder_path = Path(folder_path)
        if source_dataset is None and not self.folder_path.exists():
            raise FileNotFoundError(f"Folder does not exist: {self.folder_path}")

        self.sampling_rate = sampling_rate
        self.signal_length = signal_length
        self.file_extensions = file_extensions or [".wav"]
        self._recursive = recursive
        self._lazy_loading = lazy_loading

        # Changed to a list of LazyFrame
        self._lazy_frames: list[LazyFrame[F]] = []

        self._source_dataset = source_dataset
        self._transform = transform

        if self._source_dataset:
            self._initialize_from_source()
        else:
            self._initialize_from_folder()

    def _initialize_from_source(self) -> None:
        """Initialize from a source dataset."""
        if self._source_dataset is None:
            return

        # Copy file paths from source
        file_paths = self._source_dataset._get_file_paths()
        self._lazy_frames = [LazyFrame(file_path) for file_path in file_paths]

        # Inherit other properties
        self.sampling_rate = self.sampling_rate or self._source_dataset.sampling_rate
        self.signal_length = self.signal_length or self._source_dataset.signal_length
        self.file_extensions = (
            self.file_extensions or self._source_dataset.file_extensions
        )
        self._recursive = self._source_dataset._recursive
        self.folder_path = self._source_dataset.folder_path

    def _initialize_from_folder(self) -> None:
        """Initialize from a folder."""
        self._discover_files()
        if not self._lazy_loading:
            self._load_all_files()

    def _discover_files(self) -> None:
        """Discover files in the folder and store them in a list of LazyFrame."""
        file_paths = []
        for ext in self.file_extensions:
            pattern = f"**/*{ext}" if self._recursive else f"*{ext}"
            file_paths.extend(
                sorted(p for p in self.folder_path.glob(pattern) if p.is_file())
            )

        # Remove duplicates and sort
        file_paths = sorted(list(set(file_paths)))

        # Create a list of LazyFrame
        self._lazy_frames = [LazyFrame(file_path) for file_path in file_paths]

    def _load_all_files(self) -> None:
        """Load all files."""
        for i in tqdm(range(len(self._lazy_frames)), desc="Loading/transforming"):
            try:
                self._ensure_loaded(i)
            except Exception as e:
                filepath = self._lazy_frames[i].file_path
                logger.warning(
                    f"Failed to load/transform index {i} ({filepath}): {str(e)}"
                )
        self._lazy_loading = False

    @abstractmethod
    def _load_file(self, file_path: Path) -> Optional[F]:
        """Abstract method to load a frame from a file."""
        pass

    def _load_from_source(self, index: int) -> Optional[F]:
        """Load a frame from the source dataset and transform it if necessary."""
        if self._source_dataset is None or self._transform is None:
            return None

        source_frame = self._source_dataset._ensure_loaded(index)
        if source_frame is None:
            return None

        try:
            return self._transform(source_frame)
        except Exception as e:
            logger.warning(f"Failed to transform index {index}: {str(e)}")
            return None

    def _ensure_loaded(self, index: int) -> Optional[F]:
        """Ensure the frame at the given index is loaded."""
        if not (0 <= index < len(self._lazy_frames)):
            raise IndexError(
                f"Index {index} is out of range (0-{len(self._lazy_frames) - 1})"
            )

        lazy_frame = self._lazy_frames[index]

        # Return if already loaded
        if lazy_frame.is_loaded:
            return lazy_frame.frame

        try:
            # Convert from source dataset
            if self._transform and self._source_dataset:
                lazy_frame.load_attempted = True
                frame = self._load_from_source(index)
                lazy_frame.frame = frame
                lazy_frame.is_loaded = True
                return frame
            # Load directly from file
            else:
                return lazy_frame.ensure_loaded(self._load_file)
        except Exception as e:
            f_path = lazy_frame.file_path
            logger.error(
                f"Failed to load or initialize index {index} ({f_path}): {str(e)}"
            )
            lazy_frame.frame = None
            lazy_frame.is_loaded = True
            lazy_frame.load_attempted = True
            return None

    def _get_file_paths(self) -> list[Path]:
        """Get a list of file paths."""
        return [lazy_frame.file_path for lazy_frame in self._lazy_frames]

    def __len__(self) -> int:
        """Return the number of files in the dataset."""
        return len(self._lazy_frames)

    def __getitem__(self, index: int) -> Optional[F]:
        """Get the frame at the specified index."""
        return self._ensure_loaded(index)

    @overload
    def apply(self, func: Callable[[F], Optional[F_out]]) -> "FrameDataset[F_out]": ...

    @overload
    def apply(self, func: Callable[[F], Optional[Any]]) -> "FrameDataset[Any]": ...

    def apply(self, func: Callable[[F], Optional[Any]]) -> "FrameDataset[Any]":
        """Apply a function to the entire dataset to create a new dataset."""
        new_dataset = type(self)(
            folder_path=str(self.folder_path),
            lazy_loading=True,
            source_dataset=self,
            transform=func,
            sampling_rate=self.sampling_rate,
            signal_length=self.signal_length,
            file_extensions=self.file_extensions,
            recursive=self._recursive,
        )
        return cast("FrameDataset[Any]", new_dataset)

    def save(self, output_folder: str, filename_prefix: str = "") -> None:
        """Save processed frames to files."""
        raise NotImplementedError("The save method is not currently implemented.")

    def sample(
        self,
        n: Optional[int] = None,
        ratio: Optional[float] = None,
        seed: Optional[int] = None,
    ) -> "FrameDataset[F]":
        """Get a sample from the dataset."""
        if seed is not None:
            random.seed(seed)

        total = len(self._lazy_frames)
        if total == 0:
            return type(self)(
                str(self.folder_path),
                sampling_rate=self.sampling_rate,
                signal_length=self.signal_length,
                file_extensions=self.file_extensions,
                lazy_loading=self._lazy_loading,
                recursive=self._recursive,
            )

        # Determine sample size
        if n is None and ratio is None:
            n = max(1, min(10, int(total * 0.1)))
        elif n is None and ratio is not None:
            n = max(1, int(total * ratio))
        elif n is not None:
            n = max(1, n)
        else:
            n = 1

        n = min(n, total)

        # Randomly select indices
        sampled_indices = sorted(random.sample(range(total), n))

        return _SampledFrameDataset(self, sampled_indices)

    def get_metadata(self) -> dict[str, Any]:
        """Get metadata for the dataset."""
        actual_sr: Optional[Union[int, float]] = self.sampling_rate
        frame_type_name = "Unknown"

        # Count loaded frames
        loaded_count = sum(
            1 for lazy_frame in self._lazy_frames if lazy_frame.is_loaded
        )

        # Get metadata from the first frame (if possible)
        first_frame: Optional[F] = None
        if len(self._lazy_frames) > 0:
            try:
                if self._lazy_frames[0].is_loaded:
                    first_frame = self._lazy_frames[0].frame

                if first_frame:
                    actual_sr = getattr(
                        first_frame, "sampling_rate", self.sampling_rate
                    )
                    frame_type_name = type(first_frame).__name__
            except Exception as e:
                logger.warning(
                    f"Error accessing the first frame during metadata retrieval: {e}"
                )

        return {
            "folder_path": str(self.folder_path),
            "file_count": len(self._lazy_frames),
            "loaded_count": loaded_count,
            "target_sampling_rate": self.sampling_rate,
            "actual_sampling_rate": actual_sr,
            "signal_length": self.signal_length,
            "file_extensions": self.file_extensions,
            "lazy_loading": self._lazy_loading,
            "recursive": self._recursive,
            "frame_type": frame_type_name,
            "has_transform": self._transform is not None,
            "is_sampled": isinstance(self, _SampledFrameDataset),
        }


class _SampledFrameDataset(FrameDataset[F]):
    """
    A class representing a subset of a dataset.
    Contains only the indices selected from the original dataset.
    """

    def __init__(
        self,
        original_dataset: "FrameDataset[F]",
        sampled_indices: list[int],
    ):
        """
        Initialize a sampled dataset.

        Args:
            original_dataset: The original dataset
            sampled_indices: List of selected indices
        """
        # Initialize base class
        super().__init__(
            folder_path=str(original_dataset.folder_path),
            lazy_loading=True,  # Sampled datasets always use lazy loading
            sampling_rate=original_dataset.sampling_rate,
            signal_length=original_dataset.signal_length,
            file_extensions=original_dataset.file_extensions,
            recursive=original_dataset._recursive,
        )

        # Store the original dataset
        self._original_dataset = original_dataset

        # Mapping of sampled indices
        self._original_indices = sampled_indices

        # Get file paths from the original dataset and create new LazyFrames
        original_file_paths = original_dataset._get_file_paths()
        try:
            sampled_file_paths = [original_file_paths[i] for i in sampled_indices]
            self._lazy_frames = [
                LazyFrame(file_path) for file_path in sampled_file_paths
            ]
        except IndexError as e:
            logger.error("Sampled indices are out of range for the original dataset")
            logger.error(f"  Original dataset file count: {len(original_file_paths)}")
            logger.error(f"  Sampled indices: {sampled_indices}")
            raise IndexError(
                "Indices are out of range for the original dataset. Original dataset count: "  # noqa: E501
                f"{len(original_file_paths)}, indices: {sampled_indices}"
            ) from e

    def _load_file(self, file_path: Path) -> Optional[F]:
        """This class does not load directly from files but from the original dataset."""  # noqa: E501
        raise NotImplementedError("_SampledFrameDataset does not load files directly.")

    def _ensure_loaded(self, index: int) -> Optional[F]:
        """
        Load the frame corresponding to the index in the sampled dataset.
        Get the frame from the original dataset.
        """
        # Check index range
        if not (0 <= index < len(self._lazy_frames)):
            raise IndexError(
                f"Index {index} is out of range for the sampled dataset "
                f"(0-{len(self._lazy_frames) - 1})"
            )

        lazy_frame = self._lazy_frames[index]

        # Return from cache if already loaded
        if lazy_frame.is_loaded:
            return lazy_frame.frame

        # Index in the original dataset
        original_index = self._original_indices[index]

        try:
            # Get frame from the original dataset
            frame = self._original_dataset[original_index]

            # Update LazyFrame
            lazy_frame.frame = frame
            lazy_frame.is_loaded = True
            lazy_frame.load_attempted = True

            return frame

        except Exception as e:
            logger.error(
                "Error loading frame in sampled dataset "
                f"(index {index}, original index {original_index}): {str(e)}"
            )
            lazy_frame.frame = None
            lazy_frame.is_loaded = True
            lazy_frame.load_attempted = True
            return None

    @overload
    def apply(self, func: Callable[[F], Optional[F_out]]) -> "FrameDataset[F_out]": ...

    @overload
    def apply(self, func: Callable[[F], Optional[Any]]) -> "FrameDataset[Any]": ...

    def apply(self, func: Callable[[F], Optional[Any]]) -> "FrameDataset[Any]":
        """
        Apply a function to the entire sampled dataset.
        Added as a new transformation to the original dataset, maintaining sampling.
        """
        # Apply transformation to the original dataset
        transformed_dataset = self._original_dataset.apply(func)

        # Create a new sampled dataset with the same sampling indices
        return _SampledFrameDataset(transformed_dataset, self._original_indices)


class ChannelFrameDataset(FrameDataset[ChannelFrame]):
    """
    Dataset class for handling audio files as ChannelFrames in a folder.
    """

    def __init__(
        self,
        folder_path: str,
        sampling_rate: Optional[int] = None,
        signal_length: Optional[int] = None,
        file_extensions: Optional[list[str]] = None,
        lazy_loading: bool = True,
        recursive: bool = False,
        source_dataset: Optional["FrameDataset[Any]"] = None,
        transform: Optional[Callable[[Any], Optional[ChannelFrame]]] = None,
    ):
        _file_extensions = file_extensions or [
            ".wav",
            ".mp3",
            ".flac",
            ".csv",
        ]

        super().__init__(
            folder_path=folder_path,
            sampling_rate=sampling_rate,
            signal_length=signal_length,
            file_extensions=_file_extensions,
            lazy_loading=lazy_loading,
            recursive=recursive,
            source_dataset=source_dataset,
            transform=transform,
        )

    def _load_file(self, file_path: Path) -> Optional[ChannelFrame]:
        """Load an audio file and return a ChannelFrame."""
        try:
            frame = ChannelFrame.from_file(file_path)
            if self.sampling_rate and frame.sampling_rate != self.sampling_rate:
                logger.info(
                    f"Resampling file {file_path.name} ({frame.sampling_rate} Hz) to "
                    f"dataset rate ({self.sampling_rate} Hz)."
                )
                frame = frame.resampling(target_sr=self.sampling_rate)
            return frame
        except Exception as e:
            logger.error(f"Failed to load or initialize file {file_path}: {str(e)}")
            return None

    def resample(self, target_sr: int) -> "ChannelFrameDataset":
        """Resample all frames in the dataset."""

        def _resample_func(frame: ChannelFrame) -> Optional[ChannelFrame]:
            if frame is None:
                return None
            try:
                return frame.resampling(target_sr=target_sr)
            except Exception as e:
                logger.warning(f"Resampling error (target_sr={target_sr}): {e}")
                return None

        new_dataset = self.apply(_resample_func)
        return cast(ChannelFrameDataset, new_dataset)

    def trim(self, start: float, end: float) -> "ChannelFrameDataset":
        """Trim all frames in the dataset."""

        def _trim_func(frame: ChannelFrame) -> Optional[ChannelFrame]:
            if frame is None:
                return None
            try:
                return frame.trim(start=start, end=end)
            except Exception as e:
                logger.warning(f"Trimming error (start={start}, end={end}): {e}")
                return None

        new_dataset = self.apply(_trim_func)
        return cast(ChannelFrameDataset, new_dataset)

    def normalize(self, **kwargs: Any) -> "ChannelFrameDataset":
        """Normalize all frames in the dataset."""

        def _normalize_func(frame: ChannelFrame) -> Optional[ChannelFrame]:
            if frame is None:
                return None
            try:
                return frame.normalize(**kwargs)
            except Exception as e:
                logger.warning(f"Normalization error ({kwargs}): {e}")
                return None

        new_dataset = self.apply(_normalize_func)
        return cast(ChannelFrameDataset, new_dataset)

    def stft(
        self,
        n_fft: int = 2048,
        hop_length: Optional[int] = None,
        win_length: Optional[int] = None,
        window: str = "hann",
    ) -> "SpectrogramFrameDataset":
        """Apply STFT to all frames in the dataset."""
        _hop = hop_length or n_fft // 4

        def _stft_func(frame: ChannelFrame) -> Optional[SpectrogramFrame]:
            if frame is None:
                return None
            try:
                return frame.stft(
                    n_fft=n_fft,
                    hop_length=_hop,
                    win_length=win_length,
                    window=window,
                )
            except Exception as e:
                logger.warning(f"STFT error (n_fft={n_fft}, hop={_hop}): {e}")
                return None

        new_dataset = SpectrogramFrameDataset(
            folder_path=str(self.folder_path),
            lazy_loading=True,
            source_dataset=self,
            transform=_stft_func,
            sampling_rate=self.sampling_rate,
        )
        return new_dataset

    @classmethod
    def from_folder(
        cls,
        folder_path: str,
        sampling_rate: Optional[int] = None,
        file_extensions: Optional[list[str]] = None,
        recursive: bool = False,
        lazy_loading: bool = True,
    ) -> "ChannelFrameDataset":
        """Class method to create a ChannelFrameDataset from a folder."""
        extensions = (
            file_extensions
            if file_extensions is not None
            else [".wav", ".mp3", ".flac", ".csv"]
        )

        return cls(
            folder_path,
            sampling_rate=sampling_rate,
            file_extensions=extensions,
            lazy_loading=lazy_loading,
            recursive=recursive,
        )


class SpectrogramFrameDataset(FrameDataset[SpectrogramFrame]):
    """
    Dataset class for handling spectrogram data as SpectrogramFrames.
    Expected to be generated mainly as a result of ChannelFrameDataset.stft().
    """

    def __init__(
        self,
        folder_path: str,
        sampling_rate: Optional[int] = None,
        signal_length: Optional[int] = None,
        file_extensions: Optional[list[str]] = None,
        lazy_loading: bool = True,
        recursive: bool = False,
        source_dataset: Optional["FrameDataset[Any]"] = None,
        transform: Optional[Callable[[Any], Optional[SpectrogramFrame]]] = None,
    ):
        super().__init__(
            folder_path=folder_path,
            sampling_rate=sampling_rate,
            signal_length=signal_length,
            file_extensions=file_extensions,
            lazy_loading=lazy_loading,
            recursive=recursive,
            source_dataset=source_dataset,
            transform=transform,
        )

    def _load_file(self, file_path: Path) -> Optional[SpectrogramFrame]:
        """Direct loading from files is not currently supported."""
        logger.warning(
            "No method defined for directly loading SpectrogramFrames. Normally "
            "created from ChannelFrameDataset.stft()."
        )
        raise NotImplementedError(
            "No method defined for directly loading SpectrogramFrames"
        )

    def plot(self, index: int, **kwargs: Any) -> None:
        """Plot the spectrogram at the specified index."""
        try:
            frame = self._ensure_loaded(index)

            if frame is None:
                logger.warning(
                    f"Cannot plot index {index} as it failed to load/transform."
                )
                return

            plot_method = getattr(frame, "plot", None)
            if callable(plot_method):
                plot_method(**kwargs)
            else:
                logger.warning(
                    f"Frame (index {index}, type {type(frame).__name__}) does not "
                    f"have a plot method implemented."
                )
        except Exception as e:
            logger.error(f"An error occurred while plotting index {index}: {e}")



"""introspection utilities module tests."""

from wandas.utils.introspection import accepted_kwargs, filter_kwargs


def test_accepted_kwargs_with_no_kwargs():
    """Test accepted_kwargs with a function that has no kwargs."""

    def func(a, b, c):
        return a + b + c

    params, has_var_kwargs = accepted_kwargs(func)
    assert params == {"a", "b", "c"}
    assert not has_var_kwargs


def test_accepted_kwargs_with_var_kwargs():
    """Test accepted_kwargs with a function that has var kwargs."""

    def func(a, b, **kwargs):
        return a + b + sum(kwargs.values())

    params, has_var_kwargs = accepted_kwargs(func)
    assert params == {"a", "b"}
    assert has_var_kwargs


def test_filter_kwargs_with_no_var_kwargs():
    """Test filter_kwargs with a function that has no var kwargs."""

    def func(a, b, c=1):
        return a + b + c

    kwargs = {"a": 1, "b": 2, "c": 3, "d": 4}
    filtered = filter_kwargs(func, kwargs)
    assert filtered == {"a": 1, "b": 2, "c": 3}
    assert "d" not in filtered


def test_filter_kwargs_with_var_kwargs():
    """Test filter_kwargs with a function that has var kwargs."""

    def func(a, b, **kwargs):
        return a + b + sum(kwargs.values())

    kwargs = {"a": 1, "b": 2, "c": 3, "d": 4}
    filtered = filter_kwargs(func, kwargs)
    assert filtered == kwargs


def test_filter_kwargs_with_var_kwargs_strict_mode():
    """Test filter_kwargs with a function that has var kwargs in strict mode."""

    def func(a, b, **kwargs):
        return a + b + sum(kwargs.values())

    kwargs = {"a": 1, "b": 2, "c": 3, "d": 4}
    filtered = filter_kwargs(func, kwargs, strict_mode=True)
    assert filtered == {"a": 1, "b": 2}
    assert "c" not in filtered
    assert "d" not in filtered


def test_filter_kwargs_caching():
    """Test the caching behavior of filter_kwargs."""

    def func(a, b, c=1):
        return a + b + c

    # First call to accepted_kwargs initializes the cache
    cached_result1 = accepted_kwargs(func)
    expected_params = {"a", "b", "c"}
    expected_has_var_kwargs = False
    assert cached_result1[0] == expected_params
    assert cached_result1[1] is expected_has_var_kwargs

    # Subsequent calls should return the same values
    cached_result2 = accepted_kwargs(func)
    assert cached_result2[0] == cached_result1[0]
    assert cached_result2[1] is cached_result1[1]



import logging
from collections.abc import Iterator
from pathlib import Path
from typing import Any, Callable, Optional, TypeVar, Union

import dask
import dask.array as da
import matplotlib.pyplot as plt
import numpy as np
from dask.array.core import Array as DaskArray
from dask.array.core import concatenate, from_array
from IPython.display import Audio, display
from matplotlib.axes import Axes

from wandas.utils.types import NDArrayReal

from ..core.base_frame import BaseFrame
from ..core.metadata import ChannelMetadata
from ..io.readers import get_file_reader
from .mixins import ChannelProcessingMixin, ChannelTransformMixin

logger = logging.getLogger(__name__)

dask_delayed = dask.delayed  # type: ignore [unused-ignore]
da_from_delayed = da.from_delayed  # type: ignore [unused-ignore]
da_from_array = da.from_array  # type: ignore [unused-ignore]


S = TypeVar("S", bound="BaseFrame[Any]")


class ChannelFrame(
    BaseFrame[NDArrayReal], ChannelProcessingMixin, ChannelTransformMixin
):
    """Channel-based data frame for handling audio signals and time series data.

    This frame represents channel-based data such as audio signals and time series data,
    with each channel containing data samples in the time domain.
    """

    def __init__(
        self,
        data: DaskArray,
        sampling_rate: float,
        label: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        operation_history: Optional[list[dict[str, Any]]] = None,
        channel_metadata: Optional[list[ChannelMetadata]] = None,
        previous: Optional["BaseFrame[Any]"] = None,
    ) -> None:
        """Initialize a ChannelFrame.

        Args:
            data: Dask array containing channel data.
            Shape should be (n_channels, n_samples).
            sampling_rate: The sampling rate of the data in Hz.
            label: A label for the frame.
            metadata: Optional metadata dictionary.
            operation_history: History of operations applied to the frame.
            channel_metadata: Metadata for each channel.
            previous: Reference to the previous frame in the processing chain.

        Raises:
            ValueError: If data has more than 2 dimensions.
        """
        if data.ndim == 1:
            data = da.reshape(data, (1, -1))
        elif data.ndim > 2:
            raise ValueError(
                f"Data must be 1-dimensional or 2-dimensional. Shape: {data.shape}"
            )
        super().__init__(
            data=data,
            sampling_rate=sampling_rate,
            label=label,
            metadata=metadata,
            operation_history=operation_history,
            channel_metadata=channel_metadata,
            previous=previous,
        )

    @property
    def _n_channels(self) -> int:
        """Returns the number of channels."""
        return self.shape[-2]

    @property
    def time(self) -> NDArrayReal:
        """Get time array for the signal.

        Returns:
            Array of time points in seconds.
        """
        return np.arange(self.n_samples) / self.sampling_rate

    @property
    def n_samples(self) -> int:
        """Returns the number of samples."""
        n: int = self._data.shape[-1]
        return n

    @property
    def duration(self) -> float:
        """Returns the duration in seconds."""
        return self.n_samples / self.sampling_rate

    def _apply_operation_impl(self: S, operation_name: str, **params: Any) -> S:
        logger.debug(f"Applying operation={operation_name} with params={params} (lazy)")
        from ..processing import create_operation

        # Create operation instance
        operation = create_operation(operation_name, self.sampling_rate, **params)

        # Apply processing to data
        processed_data = operation.process(self._data)

        # Update metadata
        operation_metadata = {"operation": operation_name, "params": params}
        new_history = self.operation_history.copy()
        new_history.append(operation_metadata)
        new_metadata = {**self.metadata}
        new_metadata[operation_name] = params

        logger.debug(
            f"Created new ChannelFrame with operation {operation_name} added to graph"
        )
        if operation_name == "resampling":
            # For resampling, update sampling rate
            return self._create_new_instance(
                sampling_rate=params["target_sr"],
                data=processed_data,
                metadata=new_metadata,
                operation_history=new_history,
            )
        return self._create_new_instance(
            data=processed_data,
            metadata=new_metadata,
            operation_history=new_history,
        )

    def _binary_op(
        self,
        other: Union["ChannelFrame", int, float, NDArrayReal, "DaskArray"],
        op: Callable[["DaskArray", Any], "DaskArray"],
        symbol: str,
    ) -> "ChannelFrame":
        """
        Common implementation for binary operations
        - utilizing dask's lazy evaluation.

        Args:
            other: Right operand for the operation.
            op: Function to execute the operation (e.g., lambda a, b: a + b).
            symbol: Symbolic representation of the operation (e.g., '+').

        Returns:
            A new channel containing the operation result (lazy execution).
        """
        from .channel import ChannelFrame

        logger.debug(f"Setting up {symbol} operation (lazy)")

        # Handle potentially None metadata and operation_history
        metadata = {}
        if self.metadata is not None:
            metadata = self.metadata.copy()

        operation_history = []
        if self.operation_history is not None:
            operation_history = self.operation_history.copy()

        # Check if other is a ChannelFrame - improved type checking
        if isinstance(other, ChannelFrame):
            if self.sampling_rate != other.sampling_rate:
                raise ValueError(
                    "Sampling rates do not match. Cannot perform operation."
                )

            # Perform operation directly on dask array (maintaining lazy execution)
            result_data = op(self._data, other._data)

            # Merge channel metadata
            merged_channel_metadata = []
            for self_ch, other_ch in zip(
                self._channel_metadata, other._channel_metadata
            ):
                ch = self_ch.model_copy(deep=True)
                ch["label"] = f"({self_ch['label']} {symbol} {other_ch['label']})"
                merged_channel_metadata.append(ch)

            operation_history.append({"operation": symbol, "with": other.label})

            return ChannelFrame(
                data=result_data,
                sampling_rate=self.sampling_rate,
                label=f"({self.label} {symbol} {other.label})",
                metadata=metadata,
                operation_history=operation_history,
                channel_metadata=merged_channel_metadata,
                previous=self,
            )

        # Perform operation with scalar, NumPy array, or other types
        else:
            # Apply operation directly on dask array (maintaining lazy execution)
            result_data = op(self._data, other)

            # Operand display string
            if isinstance(other, (int, float)):
                other_str = str(other)
            elif isinstance(other, np.ndarray):
                other_str = f"ndarray{other.shape}"
            elif hasattr(other, "shape"):  # Check for dask.array.Array
                other_str = f"dask.array{other.shape}"
            else:
                other_str = str(type(other).__name__)

            # Update channel metadata
            updated_channel_metadata: list[ChannelMetadata] = []
            for self_ch in self._channel_metadata:
                ch = self_ch.model_copy(deep=True)
                ch["label"] = f"({self_ch.label} {symbol} {other_str})"
                updated_channel_metadata.append(ch)

            operation_history.append({"operation": symbol, "with": other_str})

            return ChannelFrame(
                data=result_data,
                sampling_rate=self.sampling_rate,
                label=f"({self.label} {symbol} {other_str})",
                metadata=metadata,
                operation_history=operation_history,
                channel_metadata=updated_channel_metadata,
                previous=self,
            )

    def add(
        self,
        other: Union["ChannelFrame", int, float, NDArrayReal],
        snr: Optional[float] = None,
    ) -> "ChannelFrame":
        """Add another signal or value to the current signal.

        If SNR is specified, performs addition with consideration for
        signal-to-noise ratio.

        Args:
            other: Signal or value to add.
            snr: Signal-to-noise ratio (dB). If specified, adjusts the scale of the
                other signal based on this SNR.
                self is treated as the signal, and other as the noise.

        Returns:
            A new channel frame containing the addition result (lazy execution).
        """
        logger.debug(f"Setting up add operation with SNR={snr} (lazy)")

        if isinstance(other, ChannelFrame):
            # Check if sampling rates match
            if self.sampling_rate != other.sampling_rate:
                raise ValueError(
                    "Sampling rates do not match. Cannot perform operation."
                )

        elif isinstance(other, np.ndarray):
            other = ChannelFrame.from_numpy(
                other, self.sampling_rate, label="array_data"
            )
        elif isinstance(other, (int, float)):
            return self + other
        else:
            raise TypeError(
                "Addition target with SNR must be a ChannelFrame or "
                f"NumPy array: {type(other)}"
            )

        # If SNR is specified, adjust the length of the other signal
        if other.duration != self.duration:
            other = other.fix_length(length=self.n_samples)

        if snr is None:
            return self + other
        return self.apply_operation("add_with_snr", other=other._data, snr=snr)

    def plot(
        self, plot_type: str = "waveform", ax: Optional["Axes"] = None, **kwargs: Any
    ) -> Union["Axes", Iterator["Axes"]]:
        """Plot the frame data.

        Args:
            plot_type: Type of plot. Default is "waveform".
            ax: Optional matplotlib axes for plotting.
            **kwargs: Additional arguments passed to the plot function.

        Returns:
            Single Axes object or iterator of Axes objects.
        """
        logger.debug(f"Plotting audio with plot_type={plot_type} (will compute now)")

        # Get plot strategy
        from ..visualization.plotting import create_operation

        plot_strategy = create_operation(plot_type)

        # Execute plot
        _ax = plot_strategy.plot(self, ax=ax, **kwargs)

        logger.debug("Plot rendering complete")

        return _ax

    def rms_plot(
        self,
        ax: Optional["Axes"] = None,
        title: Optional[str] = None,
        overlay: bool = True,
        Aw: bool = False,  # noqa: N803
        **kwargs: Any,
    ) -> Union["Axes", Iterator["Axes"]]:
        """Generate an RMS plot.

        Args:
            ax: Optional matplotlib axes for plotting.
            title: Title for the plot.
            overlay: Whether to overlay the plot on the existing axis.
            Aw: Apply A-weighting.
            **kwargs: Additional arguments passed to the plot function.

        Returns:
            Single Axes object or iterator of Axes objects.
        """
        kwargs = kwargs or {}
        ylabel = kwargs.pop("ylabel", "RMS")
        rms_ch: ChannelFrame = self.rms_trend(Aw=Aw, dB=True)
        return rms_ch.plot(ax=ax, ylabel=ylabel, title=title, overlay=overlay, **kwargs)

    def describe(
        self, normalize: bool = True, is_close: bool = True, **kwargs: Any
    ) -> None:
        """Display visual and audio representation of the frame.

        Args:
            normalize: Whether to normalize the audio data for playback.
            is_close: Whether to close the figure after displaying.
            **kwargs: Additional parameters for visualization.
        """
        if "axis_config" in kwargs:
            logger.warning(
                "axis_config is retained for backward compatibility but will be deprecated in the future."  # noqa: E501
            )
            axis_config = kwargs["axis_config"]
            if "time_plot" in axis_config:
                kwargs["waveform"] = axis_config["time_plot"]
            if "freq_plot" in axis_config:
                if "xlim" in axis_config["freq_plot"]:
                    vlim = axis_config["freq_plot"]["xlim"]
                    kwargs["vmin"] = vlim[0]
                    kwargs["vmax"] = vlim[1]
                if "ylim" in axis_config["freq_plot"]:
                    ylim = axis_config["freq_plot"]["ylim"]
                    kwargs["ylim"] = ylim

        if "cbar_config" in kwargs:
            logger.warning(
                "cbar_config is retained for backward compatibility but will be deprecated in the future."  # noqa: E501
            )
            cbar_config = kwargs["cbar_config"]
            if "vmin" in cbar_config:
                kwargs["vmin"] = cbar_config["vmin"]
            if "vmax" in cbar_config:
                kwargs["vmax"] = cbar_config["vmax"]

        for ch in self:
            ax: Axes
            _ax = ch.plot("describe", title=f"{ch.label} {ch.labels[0]}", **kwargs)
            if isinstance(_ax, Iterator):
                ax = next(iter(_ax))
            elif isinstance(_ax, Axes):
                ax = _ax
            else:
                raise TypeError(
                    f"Unexpected type for plot result: {type(_ax)}. Expected Axes or Iterator[Axes]."  # noqa: E501
                )
            # display関数とAudioクラスを使用
            display(ax.figure)
            if is_close:
                plt.close(ax.figure)
            display(Audio(ch.data, rate=ch.sampling_rate, normalize=normalize))

    @classmethod
    def from_numpy(
        cls,
        data: NDArrayReal,
        sampling_rate: float,
        label: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        ch_labels: Optional[list[str]] = None,
        ch_units: Optional[Union[list[str], str]] = None,
    ) -> "ChannelFrame":
        """Create a ChannelFrame from a NumPy array.

        Args:
            data: NumPy array containing channel data.
            sampling_rate: The sampling rate in Hz.
            label: A label for the frame.
            metadata: Optional metadata dictionary.
            ch_labels: Labels for each channel.
            ch_units: Units for each channel.

        Returns:
            A new ChannelFrame containing the NumPy data.
        """
        if data.ndim == 1:
            data = data.reshape(1, -1)
        elif data.ndim > 2:
            raise ValueError(
                f"Data must be 1-dimensional or 2-dimensional. Shape: {data.shape}"
            )

        # Convert NumPy array to dask array
        dask_data = da_from_array(data)
        cf = cls(
            data=dask_data,
            sampling_rate=sampling_rate,
            label=label or "numpy_data",
        )
        if metadata is not None:
            cf.metadata = metadata
        if ch_labels is not None:
            if len(ch_labels) != cf.n_channels:
                raise ValueError(
                    "Number of channel labels does not match the number of channels"
                )
            for i in range(len(ch_labels)):
                cf._channel_metadata[i].label = ch_labels[i]
        if ch_units is not None:
            if isinstance(ch_units, str):
                ch_units = [ch_units] * cf.n_channels

            if len(ch_units) != cf.n_channels:
                raise ValueError(
                    "Number of channel units does not match the number of channels"
                )
            for i in range(len(ch_units)):
                cf._channel_metadata[i].unit = ch_units[i]

        return cf

    @classmethod
    def from_ndarray(
        cls,
        array: NDArrayReal,
        sampling_rate: float,
        labels: Optional[list[str]] = None,
        unit: Optional[Union[list[str], str]] = None,
        frame_label: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
    ) -> "ChannelFrame":
        """Create a ChannelFrame from a NumPy array.

        This method is deprecated. Use from_numpy instead.

        Args:
            array: Signal data. Each row corresponds to a channel.
            sampling_rate: Sampling rate (Hz).
            labels: Labels for each channel.
            unit: Unit of the signal.
            frame_label: Label for the frame.
            metadata: Optional metadata dictionary.

        Returns:
            A new ChannelFrame containing the data.
        """
        # Redirect to from_numpy for compatibility
        # However, from_ndarray is deprecated
        logger.warning("from_ndarray is deprecated. Use from_numpy instead.")
        return cls.from_numpy(
            data=array,
            sampling_rate=sampling_rate,
            label=frame_label,
            metadata=metadata,
            ch_labels=labels,
            ch_units=unit,
        )

    @classmethod
    def from_file(
        cls,
        path: Union[str, Path],
        channel: Optional[Union[int, list[int]]] = None,
        start: Optional[float] = None,
        end: Optional[float] = None,
        chunk_size: Optional[int] = None,
        ch_labels: Optional[list[str]] = None,
        **kwargs: Any,
    ) -> "ChannelFrame":
        """Create a ChannelFrame from an audio file.

        Args:
            path: Path to the audio file.
            channel: Channel(s) to load.
            start: Start time in seconds.
            end: End time in seconds.
            chunk_size: Chunk size for processing.
            Specifies the splitting size for lazy processing.
            ch_labels: Labels for each channel.
            **kwargs: Additional arguments passed to the file reader.

        Returns:
            A new ChannelFrame containing the loaded audio data.

        Raises:
            ValueError: If channel specification is invalid.
            TypeError: If channel parameter type is invalid.
            FileNotFoundError: If the file doesn't exist.
        """
        from .channel import ChannelFrame

        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path}")

        # Get file reader
        reader = get_file_reader(path)

        # Get file info
        info = reader.get_file_info(path, **kwargs)
        sr = info["samplerate"]
        n_channels = info["channels"]
        n_frames = info["frames"]
        ch_labels = ch_labels or info.get("ch_labels", None)

        logger.debug(f"File info: sr={sr}, channels={n_channels}, frames={n_frames}")

        # Channel selection processing
        all_channels = list(range(n_channels))

        if channel is None:
            channels_to_load = all_channels
            logger.debug(f"Will load all channels: {channels_to_load}")
        elif isinstance(channel, int):
            if channel < 0 or channel >= n_channels:
                raise ValueError(
                    f"Channel specification is out of range: {channel} (valid range: 0-{n_channels - 1})"  # noqa: E501
                )
            channels_to_load = [channel]
            logger.debug(f"Will load single channel: {channel}")
        elif isinstance(channel, (list, tuple)):
            for ch in channel:
                if ch < 0 or ch >= n_channels:
                    raise ValueError(
                        f"Channel specification is out of range: {ch} (valid range: 0-{n_channels - 1})"  # noqa: E501
                    )
            channels_to_load = list(channel)
            logger.debug(f"Will load specific channels: {channels_to_load}")
        else:
            raise TypeError("channel must be int, list, or None")

        # Index calculation
        start_idx = 0 if start is None else max(0, int(start * sr))
        end_idx = n_frames if end is None else min(n_frames, int(end * sr))
        frames_to_read = end_idx - start_idx

        logger.debug(
            f"Setting up lazy load from file={path}, frames={frames_to_read}, "
            f"start_idx={start_idx}, end_idx={end_idx}"
        )

        # Settings for lazy loading
        expected_shape = (len(channels_to_load), frames_to_read)

        # Define the loading function using the file reader
        def _load_audio() -> NDArrayReal:
            logger.debug(">>> EXECUTING DELAYED LOAD <<<")
            # Use the reader to get audio data with parameters
            out = reader.get_data(path, channels_to_load, start_idx, frames_to_read)
            if not isinstance(out, np.ndarray):
                raise ValueError("Unexpected data type after reading file")
            return out

        logger.debug(
            f"Creating delayed dask task with expected shape: {expected_shape}"
        )

        # Create delayed operation
        delayed_data = dask_delayed(_load_audio)()
        logger.debug("Wrapping delayed function in dask array")

        # Create dask array from delayed computation
        dask_array = da_from_delayed(
            delayed_data, shape=expected_shape, dtype=np.float32
        )

        if chunk_size is not None:
            if chunk_size <= 0:
                raise ValueError("Chunk size must be a positive integer")
            logger.debug(f"Setting chunk size: {chunk_size} for sample axis")
            dask_array = dask_array.rechunk({0: -1, 1: chunk_size})

        logger.debug(
            "ChannelFrame setup complete - actual file reading will occur on compute()"  # noqa: E501
        )

        cf = ChannelFrame(
            data=dask_array,
            sampling_rate=sr,
            label=path.stem,
            metadata={
                "filename": str(path),
            },
        )
        if ch_labels is not None:
            if len(ch_labels) != len(cf):
                raise ValueError(
                    "Number of channel labels does not match the number of specified channels"  # noqa: E501
                )
            for i in range(len(ch_labels)):
                cf._channel_metadata[i].label = ch_labels[i]
        return cf

    @classmethod
    def read_wav(
        cls, filename: str, labels: Optional[list[str]] = None
    ) -> "ChannelFrame":
        """Utility method to read a WAV file.

        Args:
            filename: Path to the WAV file.
            labels: Labels to set for each channel.

        Returns:
            A new ChannelFrame containing the data (lazy loading).
        """
        from .channel import ChannelFrame

        cf = ChannelFrame.from_file(filename, ch_labels=labels)
        return cf

    @classmethod
    def read_csv(
        cls,
        filename: str,
        time_column: Union[int, str] = 0,
        labels: Optional[list[str]] = None,
        delimiter: str = ",",
        header: Optional[int] = 0,
    ) -> "ChannelFrame":
        """Utility method to read a CSV file.

        Args:
            filename: Path to the CSV file.
            time_column: Index or name of the time column.
            labels: Labels to set for each channel.
            delimiter: Delimiter character.
            header: Row number to use as header.

        Returns:
            A new ChannelFrame containing the data (lazy loading).
        """
        from .channel import ChannelFrame

        cf = ChannelFrame.from_file(
            filename,
            ch_labels=labels,
            time_column=time_column,
            delimiter=delimiter,
            header=header,
        )
        return cf

    def to_wav(self, path: Union[str, Path], format: Optional[str] = None) -> None:
        """Save the audio data to a WAV file.

        Args:
            path: Path to save the file.
            format: File format. If None, determined from file extension.
        """
        from wandas.io.wav_io import write_wav

        write_wav(str(path), self, format=format)

    def save(
        self,
        path: Union[str, Path],
        *,
        format: str = "hdf5",
        compress: Optional[str] = "gzip",
        overwrite: bool = False,
        dtype: Optional[Union[str, np.dtype[Any]]] = None,
    ) -> None:
        """Save the ChannelFrame to a WDF (Wandas Data File) format.

        This saves the complete frame including all channel data and metadata
        in a format that can be loaded back with full fidelity.

        Args:
            path: Path to save the file. '.wdf' extension will be added if not present.
            format: Format to use (currently only 'hdf5' is supported)
            compress: Compression method ('gzip' by default, None for no compression)
            overwrite: Whether to overwrite existing file
            dtype: Optional data type conversion before saving (e.g. 'float32')

        Raises:
            FileExistsError: If the file exists and overwrite=False.
            NotImplementedError: For unsupported formats.

        Example:
            >>> cf = ChannelFrame.read_wav("audio.wav")
            >>> cf.save("audio_analysis.wdf")
        """
        from ..io.wdf_io import save as wdf_save

        wdf_save(
            self,
            path,
            format=format,
            compress=compress,
            overwrite=overwrite,
            dtype=dtype,
        )

    @classmethod
    def load(cls, path: Union[str, Path], *, format: str = "hdf5") -> "ChannelFrame":
        """Load a ChannelFrame from a WDF (Wandas Data File) file.

        This loads data saved with the save() method, preserving all channel data,
        metadata, labels, and units.

        Args:
            path: Path to the WDF file
            format: Format of the file (currently only 'hdf5' is supported)

        Returns:
            A new ChannelFrame with all data and metadata loaded

        Raises:
            FileNotFoundError: If the file doesn't exist
            NotImplementedError: For unsupported formats

        Example:
            >>> cf = ChannelFrame.load("audio_analysis.wdf")
        """
        from ..io.wdf_io import load as wdf_load

        return wdf_load(path, format=format)

    def _get_additional_init_kwargs(self) -> dict[str, Any]:
        """Provide additional initialization arguments required for ChannelFrame."""
        return {}

    def add_channel(
        self,
        data: Union[np.ndarray[Any, Any], DaskArray, "ChannelFrame"],
        label: Optional[str] = None,
        align: str = "strict",
        suffix_on_dup: Optional[str] = None,
        inplace: bool = False,
        **kwargs: Any,
    ) -> "ChannelFrame":
        # ndarray/dask/同型Frame対応
        if isinstance(data, ChannelFrame):
            if self.sampling_rate != data.sampling_rate:
                raise ValueError("sampling_rate不一致")
            if data.n_samples != self.n_samples:
                if align == "pad":
                    pad_len = self.n_samples - data.n_samples
                    arr = data._data
                    if pad_len > 0:
                        arr = concatenate(
                            [
                                arr,
                                from_array(
                                    np.zeros((arr.shape[0], pad_len), dtype=arr.dtype)
                                ),
                            ],
                            axis=1,
                        )
                    else:
                        arr = arr[:, : self.n_samples]
                elif align == "truncate":
                    arr = data._data[:, : self.n_samples]
                    if arr.shape[1] < self.n_samples:
                        pad_len = self.n_samples - arr.shape[1]
                        arr = concatenate(
                            [
                                arr,
                                from_array(
                                    np.zeros((arr.shape[0], pad_len), dtype=arr.dtype)
                                ),
                            ],
                            axis=1,
                        )
                else:
                    raise ValueError("データ長不一致: align指定を確認")
            else:
                arr = data._data
            labels = [ch.label for ch in self._channel_metadata]
            new_labels = []
            for chmeta in data._channel_metadata:
                new_label = chmeta.label
                if new_label in labels or new_label in new_labels:
                    if suffix_on_dup:
                        new_label += suffix_on_dup
                    else:
                        raise ValueError(f"label重複: {new_label}")
                new_labels.append(new_label)
            new_data = concatenate([self._data, arr], axis=0)
            from ..core.metadata import ChannelMetadata

            new_chmeta = self._channel_metadata + [
                ChannelMetadata(label=lbl) for lbl in new_labels
            ]
            if inplace:
                self._data = new_data
                self._channel_metadata = new_chmeta
                return self
            else:
                return ChannelFrame(
                    data=new_data,
                    sampling_rate=self.sampling_rate,
                    label=self.label,
                    metadata=self.metadata,
                    operation_history=self.operation_history,
                    channel_metadata=new_chmeta,
                    previous=self,
                )
        if isinstance(data, np.ndarray):
            arr = from_array(data.reshape(1, -1))
        elif isinstance(data, DaskArray):
            arr = data[None, ...] if data.ndim == 1 else data
            if arr.shape[0] != 1:
                arr = arr.reshape((1, -1))
        else:
            raise TypeError("add_channel: ndarray/dask/同型Frameのみ対応")
        if arr.shape[1] != self.n_samples:
            if align == "pad":
                pad_len = self.n_samples - arr.shape[1]
                if pad_len > 0:
                    arr = concatenate(
                        [arr, from_array(np.zeros((1, pad_len), dtype=arr.dtype))],
                        axis=1,
                    )
                else:
                    arr = arr[:, : self.n_samples]
            elif align == "truncate":
                arr = arr[:, : self.n_samples]
                if arr.shape[1] < self.n_samples:
                    pad_len = self.n_samples - arr.shape[1]
                    arr = concatenate(
                        [arr, from_array(np.zeros((1, pad_len), dtype=arr.dtype))],
                        axis=1,
                    )
            else:
                raise ValueError("データ長不一致: align指定を確認")
        labels = [ch.label for ch in self._channel_metadata]
        new_label = label or f"ch{len(labels)}"
        if new_label in labels:
            if suffix_on_dup:
                new_label += suffix_on_dup
            else:
                raise ValueError("label重複")
        new_data = concatenate([self._data, arr], axis=0)
        from ..core.metadata import ChannelMetadata

        new_chmeta = self._channel_metadata + [ChannelMetadata(label=new_label)]
        if inplace:
            self._data = new_data
            self._channel_metadata = new_chmeta
            return self
        else:
            return ChannelFrame(
                data=new_data,
                sampling_rate=self.sampling_rate,
                label=self.label,
                metadata=self.metadata,
                operation_history=self.operation_history,
                channel_metadata=new_chmeta,
                previous=self,
            )

    def remove_channel(
        self, key: Union[int, str], inplace: bool = False
    ) -> "ChannelFrame":
        if isinstance(key, int):
            if not (0 <= key < self.n_channels):
                raise IndexError(f"index {key} out of range")
            idx = key
        else:
            labels = [ch.label for ch in self._channel_metadata]
            if key not in labels:
                raise KeyError(f"label {key} not found")
            idx = labels.index(key)
        new_data = self._data[[i for i in range(self.n_channels) if i != idx], :]
        new_chmeta = [ch for i, ch in enumerate(self._channel_metadata) if i != idx]
        if inplace:
            self._data = new_data
            self._channel_metadata = new_chmeta
            return self
        else:
            return ChannelFrame(
                data=new_data,
                sampling_rate=self.sampling_rate,
                label=self.label,
                metadata=self.metadata,
                operation_history=self.operation_history,
                channel_metadata=new_chmeta,
                previous=self,
            )



from typing import TYPE_CHECKING

import librosa
import numpy as np
from scipy.signal.windows import tukey

if TYPE_CHECKING:
    from wandas.utils.types import NDArrayReal


def unit_to_ref(unit: str) -> float:
    """
    Convert unit to reference value.

    Parameters
    ----------
    unit : str
        Unit string.

    Returns
    -------
    float
        Reference value for the unit. For 'Pa', returns 2e-5 (20 μPa).
        For other units, returns 1.0.
    """
    if unit == "Pa":
        return 2e-5

    else:
        return 1.0


def calculate_rms(wave: "NDArrayReal") -> "NDArrayReal":
    """
    Calculate the root mean square of the wave.

    Parameters
    ----------
    wave : NDArrayReal
        Input waveform data. Can be multi-channel (shape: [channels, samples])
        or single channel (shape: [samples]).

    Returns
    -------
    Union[float, NDArray[np.float64]]
        RMS value(s). For multi-channel input, returns an array of RMS values,
        one per channel. For single-channel input, returns a single RMS value.
    """
    # Calculate RMS considering axis (over the last dimension)
    axis_to_use = -1 if wave.ndim > 1 else None
    rms_values: NDArrayReal = np.sqrt(
        np.mean(np.square(wave), axis=axis_to_use, keepdims=True)
    )
    return rms_values


def calculate_desired_noise_rms(clean_rms: "NDArrayReal", snr: float) -> "NDArrayReal":
    """
    Calculate the desired noise RMS based on clean signal RMS and target SNR.

    Parameters
    ----------
    clean_rms : "NDArrayReal"
        RMS value(s) of the clean signal.
        Can be a single value or an array for multi-channel.
    snr : float
        Target Signal-to-Noise Ratio in dB.

    Returns
    -------
    "NDArrayReal"
        Desired noise RMS value(s) to achieve the target SNR.
    """
    a = snr / 20
    noise_rms = clean_rms / (10**a)
    return noise_rms


def amplitude_to_db(amplitude: "NDArrayReal", ref: float) -> "NDArrayReal":
    """
    Convert amplitude to decibel.

    Parameters
    ----------
    amplitude : NDArrayReal
        Input amplitude data.
    ref : float
        Reference value for conversion.

    Returns
    -------
    NDArrayReal
        Amplitude data converted to decibels.
    """
    db: NDArrayReal = librosa.amplitude_to_db(
        np.abs(amplitude), ref=ref, amin=1e-15, top_db=None
    )
    return db


def level_trigger(
    data: "NDArrayReal", level: float, offset: int = 0, hold: int = 1
) -> list[int]:
    """
    Find points where the signal crosses the specified level from below.

    Parameters
    ----------
    data : NDArrayReal
        Input signal data.
    level : float
        Threshold level for triggering.
    offset : int, default=0
        Offset to add to trigger points.
    hold : int, default=1
        Minimum number of samples between successive trigger points.

    Returns
    -------
    list of int
        List of sample indices where the signal crosses the level.
    """
    trig_point: list[int] = []

    sig_len = len(data)
    diff = np.diff(np.sign(data - level))
    level_point = np.where(diff > 0)[0]
    level_point = level_point[(level_point + hold) < sig_len]

    if len(level_point) == 0:
        return list()

    last_point = level_point[0]
    trig_point.append(last_point + offset)
    for i in level_point:
        if (last_point + hold) < i:
            trig_point.append(i + offset)
            last_point = i

    return trig_point


def cut_sig(
    data: "NDArrayReal",
    point_list: list[int],
    cut_len: int,
    taper_rate: float = 0,
    dc_cut: bool = False,
) -> "NDArrayReal":
    """
    Cut segments from signal at specified points.

    Parameters
    ----------
    data : NDArrayReal
        Input signal data.
    point_list : list of int
        List of starting points for cutting.
    cut_len : int
        Length of each segment to cut.
    taper_rate : float, default=0
        Taper rate for Tukey window applied to segments.
        A value of 0 means no tapering, 1 means full tapering.
    dc_cut : bool, default=False
        Whether to remove DC component (mean) from segments.

    Returns
    -------
    NDArrayReal
        Array containing cut segments with shape (n_segments, cut_len).
    """
    length = len(data)
    point_list_ = [p for p in point_list if p >= 0 and p + cut_len <= length]
    trial = np.zeros((len(point_list_), cut_len))

    for i, v in enumerate(point_list_):
        trial[i] = data[v : v + cut_len]
        if dc_cut:
            trial[i] = trial[i] - trial[i].mean()

    trial = trial * tukey(cut_len, taper_rate)
    return trial



import json
from typing import Any

from wandas.core.metadata import ChannelMetadata

# filepath: wandas/core/test_channel_metadata.py


class TestChannelMetadata:
    def test_init_default_values(self) -> None:
        """Test initialization with default values"""
        metadata: ChannelMetadata = ChannelMetadata()
        assert metadata.label == ""
        assert metadata.unit == ""
        assert metadata.extra == {}

    def test_init_custom_values(self) -> None:
        """Test initialization with custom values"""
        metadata: ChannelMetadata = ChannelMetadata(
            label="test_label",
            unit="Hz",
            extra={"source": "microphone", "calibrated": True},
        )
        assert metadata.label == "test_label"
        assert metadata.unit == "Hz"
        assert metadata.extra == {"source": "microphone", "calibrated": True}

    def test_getitem_main_fields(self) -> None:
        """Test dictionary-like access for main fields"""
        metadata: ChannelMetadata = ChannelMetadata(label="test_label", unit="Hz")
        assert metadata["label"] == "test_label"
        assert metadata["unit"] == "Hz"

    def test_getitem_extra_field(self) -> None:
        """Test dictionary-like access for extra fields"""
        metadata: ChannelMetadata = ChannelMetadata(
            extra={"source": "microphone", "calibrated": True}
        )
        assert metadata["source"] == "microphone"
        assert metadata["calibrated"] is True
        # Non-existent key should return None
        assert metadata["nonexistent"] is None

    def test_setitem_main_fields(self) -> None:
        """Test dictionary-like assignment for main fields"""
        metadata: ChannelMetadata = ChannelMetadata()
        metadata["label"] = "new_label"
        metadata["unit"] = "dB"
        assert metadata.label == "new_label"
        assert metadata.unit == "dB"

    def test_setitem_extra_fields(self) -> None:
        """Test dictionary-like assignment for extra fields"""
        metadata: ChannelMetadata = ChannelMetadata()
        metadata["source"] = "microphone"
        metadata["calibrated"] = True
        assert metadata.extra == {"source": "microphone", "calibrated": True}

    def test_to_json(self) -> None:
        """Test serialization to JSON"""
        metadata: ChannelMetadata = ChannelMetadata(
            label="test_label",
            unit="Hz",
            extra={"source": "microphone", "calibrated": True},
        )
        json_data: str = metadata.to_json()
        # Validate it's proper JSON
        parsed: dict[str, Any] = json.loads(json_data)
        assert parsed["label"] == "test_label"
        assert parsed["unit"] == "Hz"
        assert parsed["extra"]["source"] == "microphone"
        assert parsed["extra"]["calibrated"] is True

    def test_from_json(self) -> None:
        """Test deserialization from JSON"""
        json_data: str = """
        {
            "label": "test_label",
            "unit": "Hz",
            "extra": {
                "source": "microphone",
                "calibrated": true,
                "notes": "Test recording"
            }
        }
        """
        metadata: ChannelMetadata = ChannelMetadata.from_json(json_data)
        assert metadata.label == "test_label"
        assert metadata.unit == "Hz"
        assert metadata.extra["source"] == "microphone"
        assert metadata.extra["calibrated"] is True
        assert metadata.extra["notes"] == "Test recording"

    def test_copy(self) -> None:
        """Test deep copying of metadata"""
        metadata: ChannelMetadata = ChannelMetadata(
            label="test_label",
            unit="Hz",
            extra={"source": "microphone", "calibrated": True},
        )
        copy_mata: ChannelMetadata = metadata.copy(deep=True)

        # Verify all fields are equal
        assert copy_mata.label == metadata.label
        assert copy_mata.unit == metadata.unit
        assert copy_mata.extra == metadata.extra

        # Verify it's a deep copy by modifying the original
        metadata.label = "modified_label"
        metadata.extra["new_key"] = "new_value"

        # The copy should remain unchanged
        assert copy_mata.label == "test_label"
        assert "new_key" not in copy_mata.extra

    def test_unicode_and_special_chars(self) -> None:
        """Test handling of Unicode and special characters"""
        metadata: ChannelMetadata = ChannelMetadata(
            label="测试标签",  # Chinese characters
            unit="°C",  # Degree symbol
            extra={"note": "Special chars: !@#$%^&*()"},
        )

        # Test serialization and deserialization with special chars
        json_data: str = metadata.to_json()
        deserialized: ChannelMetadata = ChannelMetadata.from_json(json_data)

        assert deserialized.label == "测试标签"
        assert deserialized.unit == "°C"
        assert deserialized.extra["note"] == "Special chars: !@#$%^&*()"

    def test_nested_extra_data(self) -> None:
        """Test handling of nested structures in extra field"""
        nested_data: dict[str, Any] = {
            "config": {"sampling": {"rate": 44100, "bits": 24}},
            "tags": ["audio", "speech", "raw"],
        }

        metadata: ChannelMetadata = ChannelMetadata(extra=nested_data)
        json_data: str = metadata.to_json()
        deserialized: ChannelMetadata = ChannelMetadata.from_json(json_data)

        assert deserialized.extra["config"]["sampling"]["rate"] == 44100
        assert deserialized.extra["config"]["sampling"]["bits"] == 24
        assert deserialized.extra["tags"] == ["audio", "speech", "raw"]
